#!/usr/bin/env python
# -*- coding:utf-8 -*-


from .meta_data_collator import (
    DataCollatorForMetaSeq2Seq,
    DynamicSSIGenerator,
)

from .t5mlm_data_collator import (
    DataCollatorForT5MLM,
)

from .hybird_data_collator import (
    HybirdDataCollator,
)


__all__ = [
    'DataCollatorForMetaSeq2Seq',
    'DynamicSSIGenerator',
    'HybirdDataCollator',
    'DataCollatorForT5MLM',
]



import json
import os

import torch
import torch.nn.functional as F
from transformers import BertConfig, BertForTokenClassification, BertTokenizer
from collections import OrderedDict
from .BiLSTM_CRF import *

import hydra
from hydra import utils
import nltk
from nltk import word_tokenize


class BertNer(BertForTokenClassification):

    def forward(
        self, 
        input_ids, 
        token_type_ids=None, 
        attention_mask=None, 
        valid_ids=None
    ):
        sequence_output = self.bert(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask, head_mask=None)[0]
        batch_size,max_len,feat_dim = sequence_output.shape
        valid_output = torch.zeros(batch_size,max_len,feat_dim,dtype=torch.float32,device='cuda' if torch.cuda.is_available() else 'cpu')
        for i in range(batch_size):
            jj = -1
            for j in range(max_len):
                    if valid_ids[i][j].item() == 1:
                        jj += 1
                        valid_output[i][jj] = sequence_output[i][j]
        sequence_output = self.dropout(valid_output)
        logits = self.classifier(sequence_output)
        return logits

class InferNer:

    def __init__(self,model_dir: str, config, vocab_size=None, num_labels=None, word2id=None, id2label=None):
        """
        Custom Your Ner Model
        Args:
            vocab_size (`int`): (Only BiLSTM_CRF) vocabulary size used in BiLSTM_CRF
            num_labels (`int`): (Only BiLSTM_CRF)the number of Labels used in BiLSTM_CRF, such as the length of ([B-LOC, I-LOC, B-ORG, I-ORG, B-PER, I-PER, O])
            word2id (`dict`): (Only BiLSTM_CRF) Map word into index for embedding
            id2label (`dict`): (Only BiLSTM_CRF) Map index into Label for decoding
        """
        self.cfg = config
        if config.model_name == 'lstmcrf':
            assert vocab_size != None
            assert num_labels != None
            assert word2id != None
            assert id2label != None

            self.word2id = word2id
            self.id2label = id2label

            self.model = torch.load(os.path.join(utils.get_original_cwd(), self.cfg.output_dir, self.cfg.model_save_name))
        elif config.model_name == 'bert':
            self.model , self.tokenizer, self.model_config = self.load_model(model_dir)
            self.label_map = self.model_config["label_map"]
            self.max_seq_length = self.model_config["max_seq_length"]
            self.label_map = {int(k):v for k,v in self.label_map.items()}
        else:
            raise NotImplementedError(f"model type {self.cfg.model_name} not supported")

        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model = self.model.to(self.device)
        self.model.eval()

    def load_model(self, model_dir: str, model_config: str = "model_config.json"):
        model_config = os.path.join(model_dir,model_config)
        model_config = json.load(open(model_config))
        model = BertNer.from_pretrained(model_dir)
        tokenizer = BertTokenizer.from_pretrained(model_dir, do_lower_case=model_config["do_lower"])
        return model, tokenizer, model_config

    def tokenize(self, text: str):
        """ tokenize input"""
        if self.cfg.lan == "en":
            nltk.download('punkt')
            words = word_tokenize(text)
        else:
            words = list(text)
        tokens = []
        valid_positions = []
        for i,word in enumerate(words):
            token = self.tokenizer.tokenize(word)
            tokens.extend(token)
            for i in range(len(token)):
                if i == 0:
                    valid_positions.append(1)
                else:
                    valid_positions.append(0)
        return tokens, valid_positions

    def preprocess(self, text: str):
        """ preprocess """
        tokens, valid_positions = self.tokenize(text)
        ## insert "[CLS]"
        tokens.insert(0,"[CLS]")
        valid_positions.insert(0,1)
        ## insert "[SEP]"
        tokens.append("[SEP]")
        valid_positions.append(1)
        segment_ids = []
        for i in range(len(tokens)):
            segment_ids.append(0)
        input_ids = self.tokenizer.convert_tokens_to_ids(tokens)
        input_mask = [1] * len(input_ids)
        while len(input_ids) < self.max_seq_length:
            input_ids.append(0)
            input_mask.append(0)
            segment_ids.append(0)
            valid_positions.append(0)
        return input_ids,input_mask,segment_ids,valid_positions

    def predict(self, text: str):
        if self.cfg.model_name == 'lstmcrf':
            words = [[self.word2id.get(w, 0) for w in text]]
            mask = [[1] * len(words[0])]
            y_pred = self.model(torch.tensor(words).to(self.device), torch.tensor(mask).bool().to(self.device))

            """Just adjust the format of output"""
            words = [w for w in text]
            labels = [(self.id2label[id], None) for id in y_pred[0]]
        elif self.cfg.model_name == 'bert':
            input_ids,input_mask,segment_ids,valid_ids = self.preprocess(text)
            input_ids = torch.tensor([input_ids],dtype=torch.long,device=self.device)
            input_mask = torch.tensor([input_mask],dtype=torch.long,device=self.device)
            segment_ids = torch.tensor([segment_ids],dtype=torch.long,device=self.device)
            valid_ids = torch.tensor([valid_ids],dtype=torch.long,device=self.device)
            with torch.no_grad():
                logits = self.model(input_ids=input_ids, token_type_ids=segment_ids, attention_mask=input_mask, valid_ids=valid_ids)
            logits = F.softmax(logits,dim=2)
            logits_label = torch.argmax(logits,dim=2)
            logits_label = logits_label.detach().cpu().numpy().tolist()[0]

            logits_confidence = [values[label].item() for values,label in zip(logits[0],logits_label)]

            logits = []
            pos = 0
            for index,mask in enumerate(valid_ids[0]):
                if index == 0:
                    continue
                if mask == 1:
                    logits.append((logits_label[index-pos],logits_confidence[index-pos]))
                else:
                    pos += 1
            logits.pop()

            labels = [(self.label_map[label],confidence) for label,confidence in logits]
            if self.cfg.lan == "en":
                nltk.download('punkt')
                words = word_tokenize(text)
            else:
                words = list(text)
            assert len(labels) == len(words)
        else:
            raise NotImplementedError(f"model type {self.cfg.model_name} not supported")

        result = []
        for word, (label, confidence) in zip(words, labels):
            if label!='O':
                result.append((word,label))
        # tmp = []
        # tag = OrderedDict()
        # tag['PER'] = []
        # tag['LOC'] = []
        # tag['ORG'] = []
        
        # for i, (word, label) in enumerate(result):
        #     if label=='B-PER' or label=='B-LOC' or label=='B-ORG':
        #         if i==0:
        #             tmp.append(word)
        #         else:
        #             wordstype = result[i-1][1][2:]
        #             tag[wordstype].append(''.join(tmp))
        #             tmp.clear()
        #             tmp.append(word)
        #     else:
        #         tmp.append(word)
                
        #     if i==len(result)-1:
        #         if label=='B-PER' or label=='B-LOC' or label=='B-ORG':
        #             tmp.append(word)
        #         wordstype = result[i][1][2:]
        #         tag[wordstype].append(''.join(tmp))

        return result



import os
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../")))
import torch
import torch.nn as nn
import torch.nn.functional as F
from . import BasicModule
from module import Embedding, CNN

from utils import seq_len_to_mask


class PCNN(BasicModule):
    def __init__(self, cfg):
        super(PCNN, self).__init__()

        self.use_pcnn = cfg.use_pcnn
        if cfg.dim_strategy == 'cat':
            cfg.in_channels = cfg.word_dim + 2 * cfg.pos_dim
        else:
            cfg.in_channels = cfg.word_dim

        self.embedding = Embedding(cfg)
        self.cnn = CNN(cfg)
        self.fc1 = nn.Linear(len(cfg.kernel_sizes) * cfg.out_channels, cfg.intermediate)
        self.fc2 = nn.Linear(cfg.intermediate, cfg.num_attributes)
        self.dropout = nn.Dropout(cfg.dropout)

        if self.use_pcnn:
            self.fc_pcnn = nn.Linear(3 * len(cfg.kernel_sizes) * cfg.out_channels,
                                     len(cfg.kernel_sizes) * cfg.out_channels)
            self.pcnn_mask_embedding = nn.Embedding(4, 3)
            masks = torch.tensor([[0, 0, 0], [100, 0, 0], [0, 100, 0], [0, 0, 100]])
            self.pcnn_mask_embedding.weight.data.copy_(masks)
            self.pcnn_mask_embedding.weight.requires_grad = False


    def forward(self, x):
        word, lens, entity_pos, attribute_value_pos = x['word'], x['lens'], x['entity_pos'], x['attribute_value_pos']
        mask = seq_len_to_mask(lens)

        inputs = self.embedding(word, entity_pos, attribute_value_pos)
        out, out_pool = self.cnn(inputs, mask=mask)

        if self.use_pcnn:
            out = out.unsqueeze(-1)  # [B, L, Hs, 1]
            pcnn_mask = x['pcnn_mask']
            pcnn_mask = self.pcnn_mask_embedding(pcnn_mask).unsqueeze(-2)  # [B, L, 1, 3]
            out = out + pcnn_mask  # [B, L, Hs, 3]
            out = out.max(dim=1)[0] - 100  # [B, Hs, 3]
            out_pool = out.view(out.size(0), -1)  # [B, 3 * Hs]
            out_pool = F.leaky_relu(self.fc_pcnn(out_pool))  # [B, Hs]
            out_pool = self.dropout(out_pool)

        output = self.fc1(out_pool)
        output = F.leaky_relu(output)
        output = self.dropout(output)
        output = self.fc2(output)

        return output



import torch
import numpy as np
from abc import ABCMeta, abstractmethod
from sklearn.metrics import precision_recall_fscore_support


class Metric(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def reset(self):
        """
        Resets the metric to to it's initial state.
        This is called at the start of each epoch.
        """
        pass

    @abstractmethod
    def update(self, *args):
        """
        Updates the metric's state using the passed batch output.
        This is called once for each batch.
        """
        pass

    @abstractmethod
    def compute(self):
        """
        Computes the metric based on it's accumulated state.
        This is called at the end of each epoch.
        :return: the actual quantity of interest
        """
        pass


class PRMetric():
    def __init__(self):
        """
        暂时调用 sklearn 的方法
        """
        self.y_true = np.empty(0)
        self.y_pred = np.empty(0)

    def reset(self):
        """
        重置为0
        """
        self.y_true = np.empty(0)
        self.y_pred = np.empty(0)

    def update(self, y_true: torch.Tensor, y_pred: torch.Tensor):
        """
        更新tensor，保留值，取消原有梯度
        """
        y_true = y_true.cpu().detach().numpy()
        y_pred = y_pred.cpu().detach().numpy()
        y_pred = np.argmax(y_pred, axis=-1)

        self.y_true = np.append(self.y_true, y_true)
        self.y_pred = np.append(self.y_pred, y_pred)

    def compute(self):
        """
        计算acc,p,r,f1并返回
        """
        p, r, f1, _ = precision_recall_fscore_support(self.y_true, self.y_pred, average='macro', warn_for=tuple())
        _, _, acc, _ = precision_recall_fscore_support(self.y_true, self.y_pred, average='micro', warn_for=tuple())

        return acc, p, r, f1



from .constraint_decoder import *
from .data_collator import *
from .constrained_seq2seq import *
from .features import *
from .modeling_t5 import *
from .models import *
from .t5_bert_tokenizer import *
from .trainer_arguments import *


import torch
import logging
import matplotlib.pyplot as plt
from .metrics import PRMetric

logger = logging.getLogger(__name__)


def train(epoch, model, dataloader, optimizer, criterion, device, writer, cfg):
    """
    training the model.
        Args:
            epoch (int): number of training steps.
            model (class): model of training.
            dataloader (dict): dict of dataset iterator. Keys are tasknames, values are corresponding dataloaders.
            optimizer (Callable): optimizer of training.
            criterion (Callable): loss criterion of training.
            device (torch.device): device of training.
            writer (class): output to tensorboard.
            cfg: configutation of training.
        Return:
            losses[-1] : the loss of training
    """
    model.train()

    metric = PRMetric()
    losses = []

    for batch_idx, (x, y) in enumerate(dataloader, 1):
        for key, value in x.items():
            x[key] = value.to(device)
            
        y = y.to(device)

        optimizer.zero_grad()
        y_pred = model(x)

        if cfg.model_name == 'capsule':
            loss = model.loss(y_pred, y)
        else:
            loss = criterion(y_pred, y)

        loss.backward()
        optimizer.step()

        metric.update(y_true=y, y_pred=y_pred)
        losses.append(loss.item())

        data_total = len(dataloader.dataset)
        data_cal = data_total if batch_idx == len(dataloader) else batch_idx * len(y)
        if (cfg.train_log and batch_idx % cfg.log_interval == 0) or batch_idx == len(dataloader):
            # p r f1 皆为 macro，因为micro时三者相同，定义为acc
            acc, p, r, f1 = metric.compute()
            logger.info(f'Train Epoch {epoch}: [{data_cal}/{data_total} ({100. * data_cal / data_total:.0f}%)]\t'
                        f'Loss: {loss.item():.6f}')
            logger.info(f'Train Epoch {epoch}: Acc: {100. * acc:.2f}%\t'
                        f'macro metrics: [p: {p:.4f}, r:{r:.4f}, f1:{f1:.4f}]')

    if cfg.show_plot and not cfg.only_comparison_plot:
        if cfg.plot_utils == 'matplot':
            plt.plot(losses)
            plt.title(f'epoch {epoch} train loss')
            plt.show()

        if cfg.plot_utils == 'tensorboard':
            for i in range(len(losses)):
                writer.add_scalar(f'epoch_{epoch}_training_loss', losses[i], i)

    return losses[-1]


def validate(epoch, model, dataloader, criterion, device, cfg):
    """
    validating the model.
        Args:
            epoch (int): number of validating steps.
            model (class): model of validating.
            dataloader (dict): dict of dataset iterator. Keys are tasknames, values are corresponding dataloaders.
            criterion (Callable): loss criterion of validating.
            device (torch.device): device of validating.
            cfg: configutation of validating.
        Return:
            f1 : f1 score
            loss : the loss of validating
    """
    model.eval()

    metric = PRMetric()
    losses = []

    for batch_idx, (x, y) in enumerate(dataloader, 1):
        for key, value in x.items():
            x[key] = value.to(device)
        y = y.to(device)
        with torch.no_grad():
            y_pred = model(x)

            if cfg.model_name == 'capsule':
                loss = model.loss(y_pred, y)
            else:
                loss = criterion(y_pred, y)

            metric.update(y_true=y, y_pred=y_pred)
            losses.append(loss.item())

    loss = sum(losses) / len(losses)
    acc, p, r, f1 = metric.compute()
    data_total = len(dataloader.dataset)

    if epoch >= 0:
        logger.info(f'Valid Epoch {epoch}: [{data_total}/{data_total}](100%)\t Loss: {loss:.6f}')
        logger.info(f'Valid Epoch {epoch}: Acc: {100. * acc:.2f}%\tmacro metrics: [p: {p:.4f}, r:{r:.4f}, f1:{f1:.4f}]')
    else:
        logger.info(f'Test Data: [{data_total}/{data_total}](100%)\t Loss: {loss:.6f}')
        logger.info(f'Test Data: Acc: {100. * acc:.2f}%\tmacro metrics: [p: {p:.4f}, r:{r:.4f}, f1:{f1:.4f}]')

    return f1, loss



from .noiser import *
from .predict_parser import *
from .constants import *
from .dataset_processer import *
from .extraction_metrics import *
from .label_tree import *
from .record_schema import *
from .scorer import *
from .utils import *


import csv
import pickle 
import os
import logging
from tqdm import tqdm, trange
from torch.utils.data import TensorDataset
import torch.nn.functional as F
import numpy as np
import torch
from collections import OrderedDict

from transformers.utils.dummy_tokenizers_objects import BertTokenizerFast

logging.basicConfig(format = '%(asctime)s - %(levelname)s - %(name)s -   %(message)s', 
                    datefmt = '%m/%d/%Y %H:%M:%S',
                    level = logging.INFO)
logger = logging.getLogger(__name__)

import json
import re

from transformers import AutoTokenizer

keyword_files = ["keyword_train.txt", "keyword_dev.txt", "keyword_test.txt"]

def tokenize(text, tokenizer):
    # berts tokenize ways
    # tokenize the [unused12345678910]
    D = [f"[unused{i}]" for i in range(10)]
    textraw = [text]
    for delimiter in D:
        ntextraw = []
        for i in range(len(textraw)):
            t = textraw[i].split(delimiter)
            for j in range(len(t)):
                ntextraw += [t[j]]
                if j != len(t)-1:
                    ntextraw += [delimiter]
        textraw = ntextraw
    text = []
    for t in textraw:
        if t in D:
            text += [t]
        else:
            tokens = tokenizer.tokenize(t, add_special_tokens=False)
            for tok in tokens:
                text += [tok]

    for idx, t in enumerate(text):
        if idx + 3 < len(text) and t == "[" and text[idx+1] == "[UNK]" and text[idx+2] == "]":
            text = text[:idx] + ["[MASK]"] + text[idx+3:]

    return text

n_class = 1
class InputExample(object):
    """A single training/test example for simple sequence classification."""

    def __init__(self, guid, text_a, text_b=None, label=None, text_c=None, entity=None):
        """Constructs a InputExample.

        Args:
            guid: Unique id for the example.
            text_a: string. The untokenized text of the first sequence. For single
            sequence tasks, only this sequence must be specified.
            text_b: (Optional) string. The untokenized text of the second sequence.
            Only must be specified for sequence pair tasks.
            label: (Optional) string. The label of the example. This should be
            specified for train and dev examples, but not for test examples.
        """
        self.guid = guid
        self.text_a = text_a
        self.text_b = text_b
        self.text_c = text_c
        self.label = label
        self.entity = entity


class InputExampleSST2(object):
    """A single training/test example for simple sequence classification."""

    def __init__(self, guid, text_a, text_b=None, label=None, text_c=None, entity=None):
        """Constructs a InputExample.

        Args:
            guid: Unique id for the example.
            text_a: string. The untokenized text of the first sequence. For single
            sequence tasks, only this sequence must be specified.
            text_b: (Optional) string. The untokenized text of the second sequence.
            Only must be specified for sequence pair tasks.
            label: (Optional) string. The label of the example. This should be
            specified for train and dev examples, but not for test examples.
        """
        self.guid = guid
        self.text_a = text_a
        self.text_b = text_b
        self.label = label


class InputFeaturesSST2(object):
    """A single set of features of data."""

    def __init__(self, input_ids, attention_mask, token_type_ids, label_id):
        self.input_ids = input_ids
        self.attention_mask = attention_mask
        self.token_type_ids = token_type_ids
        self.label_id = label_id


class InputExampleWiki80(object):
    """A single training/test example for span pair classification."""

    def __init__(self, guid, sentence, span1, span2, ner1, ner2, label):
        self.guid = guid
        self.sentence = sentence
        self.span1 = span1
        self.span2 = span2
        self.ner1 = ner1
        self.ner2 = ner2
        self.label = label

class InputFeatures(object):
    """A single set of features of data."""

    def __init__(self, input_ids, input_mask, segment_ids, label_id, entity=None):
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids
        self.label_id = label_id
        self.entity = entity


class DataProcessor(object):
    """Base class for data converters for sequence classification data sets."""

    def get_train_examples(self, data_dir):
        """Gets a collection of `InputExample`s for the train set."""
        raise NotImplementedError()

    def get_dev_examples(self, data_dir):
        """Gets a collection of `InputExample`s for the dev set."""
        raise NotImplementedError()

    def get_labels(self):
        """Gets the list of labels for this data set."""
        raise NotImplementedError()

    @classmethod
    def _read_tsv(cls, input_file, quotechar=None):
        """Reads a tab separated value file."""
        with open(input_file, "r") as f:
            reader = csv.reader(f, delimiter="\t", quotechar=quotechar)
            lines = []
            for line in reader:
                lines.append(line)
            return lines

class Sst2Processor(DataProcessor):
    """Processor for the SST-2 data set (GLUE version)."""

    def __init__(self, data_dir, a):
        super().__init__()
        self.data_dir = data_dir

    def get_example_from_tensor_dict(self, tensor_dict):
        """See base class."""
        return InputExample(
            tensor_dict["idx"].numpy(),
            tensor_dict["sentence"].numpy().decode("utf-8"),
            None,
            str(tensor_dict["label"].numpy()),
        )

    def get_train_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "train.tsv")), "train")

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "dev.tsv")), "dev")

    def get_test_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "test.tsv")), "test")

    def get_labels(self):
        """See base class."""
        return ["0", "1"]

    def _create_examples(self, lines, set_type):
        """Creates examples for the training, dev and test sets."""
        examples = []
        text_index = 0
        for (i, line) in enumerate(lines):
            if i == 0:
                continue
            guid = "%s-%s" % (set_type, i)
            text_a = line[text_index]
            label = line[1]
            examples.append(InputExample(guid=guid, text_a=text_a, text_b=None, text_c=None, label=label))
        return examples



class relossProcessor(DataProcessor): #bert_s
    def __init__(self, data_path="data", use_prompt=False):
        def is_speaker(a):
            a = a.split()
            return len(a) == 2 and a[0] == "speaker" and a[1].isdigit()
        
        # replace the speaker with [unused] token
        def rename(d, x, y):
            d = d.replace("’","'")
            d = d.replace("im","i")
            d = d.replace("...",".")
            unused = ["[unused1]", "[unused2]"]
            a = []
            if is_speaker(x):
                a += [x]
            else:
                a += [None]
            if x != y and is_speaker(y):
                a += [y]
            else:
                a += [None]
            for i in range(len(a)):
                if a[i] is None:
                    continue
                d = d.replace(a[i] + ":", unused[i] + " :")
                if x == a[i]:
                    x = unused[i]
                if y == a[i]:
                    y = unused[i]
            return d, x, y
            
        self.D = [[], [], []]
        for sid in range(3):
            # 分成三个数据集
            with open(data_path + "/"+["train.json", "dev.json", "test.json"][sid], "r", encoding="utf8") as f:
                data = json.load(f)
            for i in range(len(data)):
                for j in range(len(data[i][1])):
                    rid = []
                    for k in range(36):
                        if k+1 in data[i][1][j]["rid"]:
                            rid += [1]
                        else:
                            rid += [0]
                    d, h, t = rename(' '.join(data[i][0]).lower(), data[i][1][j]["x"].lower(), data[i][1][j]["y"].lower())
                    prompt = f"what is the relation between {h} and {t} ? {t} is the [MASK] {h} ."
                    d = [
                        prompt + d,
                        h,
                        t,
                        rid,
                        t
                    ]
                    self.D[sid] += [d]
        logger.info(str(len(self.D[0])) + "," + str(len(self.D[1])) + "," + str(len(self.D[2])))
        
    def get_train_examples(self, data_dir):
        """See base class."""
        return self._create_examples(
                self.D[0], "train")

    def get_test_examples(self, data_dir):
        """See base class."""
        return self._create_examples(
                self.D[2], "test")

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(
                self.D[1], "dev")

    def get_labels(self):
        """See base class."""
        return [str(x) for x in range(36)]

    def _create_examples(self, data, set_type):
        """Creates examples for the training and dev sets."""
        examples = []
        for (i, d) in enumerate(data):
            guid = "%s-%s" % (set_type, i)
            examples.append(InputExample(guid=guid, text_a=data[i][0], text_b=data[i][1], label=data[i][3], text_c=data[i][2], entity=data[i][4]))
            
        return examples


class bertProcessor(DataProcessor): #bert_s
    def __init__(self, data_path="data", use_prompt=False):
        def is_speaker(a):
            a = a.split()
            return len(a) == 2 and a[0] == "speaker" and a[1].isdigit()
        
        # replace the speaker with [unused] token
        def rename(d, x, y):
            d = d.replace("’","'")
            d = d.replace("im","i")
            d = d.replace("...",".")
            unused = ["[unused1]", "[unused2]"]
            a = []
            if is_speaker(x):
                a += [x]
            else:
                a += [None]
            if x != y and is_speaker(y):
                a += [y]
            else:
                a += [None]
            for i in range(len(a)):
                if a[i] is None:
                    continue
                d = d.replace(a[i] + ":", unused[i] + " :")
                if x == a[i]:
                    x = unused[i]
                if y == a[i]:
                    y = unused[i]
            return d, x, y
            
        self.D = [[], [], []]
        for sid in range(3):
            # 分成三个数据集
            with open(data_path + "/"+["train.json", "dev.json", "test.json"][sid], "r", encoding="utf8") as f:
                data = json.load(f)
            sample_idx = 0
            for i in range(len(data)):
                for j in range(len(data[i][1])):
                    rid = []
                    for k in range(36):
                        if k+1 in data[i][1][j]["rid"]:
                            rid += [1]
                        else:
                            rid += [0]
                    d, h, t = rename(' '.join(data[i][0]).lower(), data[i][1][j]["x"].lower(), data[i][1][j]["y"].lower())
                    if use_prompt:
                        prompt = f"{h} is the [MASK] {t} ."
                    else:
                        prompt = f"what is the relation between {h} and {t} ?"
                    sample_idx += 1
                    d = [
                        prompt + d,
                        h,
                        t,
                        rid,
                    ]
                    self.D[sid] += [d]
        logger.info(str(len(self.D[0])) + "," + str(len(self.D[1])) + "," + str(len(self.D[2])))
        
    def get_train_examples(self, data_dir):
        """See base class."""
        return self._create_examples(
                self.D[0], "train")

    def get_test_examples(self, data_dir):
        """See base class."""
        return self._create_examples(
                self.D[2], "test")

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(
                self.D[1], "dev")

    def get_labels(self):
        """See base class."""
        return [str(x) for x in range(36)]

    def _create_examples(self, data, set_type):
        """Creates examples for the training and dev sets."""
        examples = []
        for (i, d) in enumerate(data):
            guid = "%s-%s" % (set_type, i)
            examples.append(InputExample(guid=guid, text_a=data[i][0], text_b=data[i][1], label=data[i][3], text_c=data[i][2]))
            
        return examples


class ptuneProcessor(DataProcessor): #bert_s
    def __init__(self, data_path="data", use_prompt=False, ptune_k=6):
        def is_speaker(a):
            a = a.split()
            return len(a) == 2 and a[0] == "speaker" and a[1].isdigit()
        
        # replace the speaker with [unused] token
        def rename(d, x, y):
            d = d.replace("’","'")
            d = d.replace("im","i")
            d = d.replace("...",".")
            unused = ["[unused1]", "[unused2]"]
            a = []
            if is_speaker(x):
                a += [x]
            else:
                a += [None]
            if x != y and is_speaker(y):
                a += [y]
            else:
                a += [None]
            for i in range(len(a)):
                if a[i] is None:
                    continue
                d = d.replace(a[i] + ":", unused[i] + " :")
                if x == a[i]:
                    x = unused[i]
                if y == a[i]:
                    y = unused[i]
            return d, x, y
            
        self.D = [[], [], []]
        """
        TODO, add new samples, every sample if there is a trigger then mask trigger and replace the origin mask with right token,
        if no trigger in the sentence, random mask a word in the sentence and replace the origin mask with the right token.
        
        """
        for sid in range(3):
            # 分成三个数据集
            with open(data_path + "/"+["train.json", "dev.json", "test.json"][sid], "r", encoding="utf8") as f:
                data = json.load(f)
            sample_idx = 0
            for i in range(len(data)):
                for j in range(len(data[i][1])):
                    rid = []
                    for k in range(36):
                        if k+1 in data[i][1][j]["rid"]:
                            rid += [1]
                        else:
                            rid += [0]
                    d, h, t = rename(' '.join(data[i][0]).lower(), data[i][1][j]["x"].lower(), data[i][1][j]["y"].lower())
                    unused_word = " ".join([f"[unused{i}]" for i in range(3, ptune_k+3)])
                    # st 3,4 ; ed 5,6
                    st = [f"[unused{i}]" for i in range(3,5)]
                    ed = [f"[unused{i}]" for i in range(5,7)]
                    # 789 as prompt
                    prompt = f"[sub] {st[0]} {h} {st[1]} [sub] [unused7] [unused8] [MASK] [unused9] [obj] {ed[0]} {t} {ed[1]} [obj]."
                    
                    # for temp_i in range(10):
                    #     d = d.replace(f"speaker {temp_i}:", f"[speaker{temp_i}]")

                    sample_idx += 1
                    sample = [
                        prompt + d,
                        h,
                        t,
                        rid,
                    ]
                    self.D[sid] += [sample]
                    # multi labels, add more data in the training set
                    if i == 0:
                        for idx,trigger in enumerate(data[i][1][j]['t']):
                            if trigger != "":
                                label_token = f"[class{data[i][1][j]['rid'][idx]+1}]"
                                prompt = prompt.replace("[MASK]", label_token)
                                # first assume the model predict the same output in the trigger, ...
                                d = d.replace(trigger, "[MASK]", 1)
                        sample = [
                            prompt + d,
                            h,
                            t,
                            rid,
                        ]
                        self.D[sid] += [sample]
        logger.info(str(len(self.D[0])) + "," + str(len(self.D[1])) + "," + str(len(self.D[2])))
        
    def get_train_examples(self, data_dir):
        """See base class."""
        return self._create_examples(
                self.D[0], "train")

    def get_test_examples(self, data_dir):
        """See base class."""
        return self._create_examples(
                self.D[2], "test")

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(
                self.D[1], "dev")

    def get_labels(self):
        """See base class."""
        return [str(x) for x in range(36)]

    def _create_examples(self, data, set_type):
        """Creates examples for the training and dev sets."""
        examples = []
        for (i, d) in enumerate(data):
            guid = "%s-%s" % (set_type, i)
            examples.append(InputExample(guid=guid, text_a=data[i][0], text_b=data[i][1], label=data[i][3], text_c=data[i][2]))
            
        return examples


class wiki80Processor(DataProcessor):
    """Processor for the TACRED data set."""
    def __init__(self, data_path, use_prompt):
        super().__init__()
        self.data_dir = data_path

    @classmethod
    def _read_json(cls, input_file):
        data = []
        with open(input_file, "r", encoding='utf-8') as reader:
            all_lines = reader.readlines()
            for line in all_lines:
                ins = eval(line)
                data.append(ins)
        return data

    def get_train_examples(self, data_dir):
        """See base class."""
        return self._create_examples(
            self._read_json(os.path.join(data_dir, "train.txt")), "train")

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(
            self._read_json(os.path.join(data_dir, "val.txt")), "dev")

    def get_test_examples(self, data_dir):
        """See base class."""
        return self._create_examples(
            self._read_json(os.path.join(data_dir, "test.txt")), "test")

    def get_labels(self, negative_label="no_relation"):
        data_dir = self.data_dir
        """See base class."""
        # if 'k-shot' in self.data_dir:
        #     data_dir = os.path.abspath(os.path.join(self.data_dir, "../.."))
        # else:
        #     data_dir = self.data_dir
        with open(os.path.join(data_dir,'rel2id.json'), "r", encoding='utf-8') as reader:
            re2id = json.load(reader)
        return re2id


    def _create_examples(self, dataset, set_type):
        """Creates examples for the training and dev sets."""
        examples = []
        for example in dataset:
            sentence = example['token']
            examples.append(InputExampleWiki80(guid=None,
                            sentence=sentence,
                            # maybe some bugs here, I don't -1
                            span1=(example['h']['pos'][0], example['h']['pos'][1]),
                            span2=(example['t']['pos'][0], example['t']['pos'][1]),
                            ner1=None,
                            ner2=None,
                            label=example['relation']))
        return examples

def convert_examples_to_features_for_loss(examples, max_seq_length, tokenizer):
    print("#examples", len(examples))
    features = []
    for (ex_index, example) in enumerate(examples):
        tokens_a = tokenize(example.text_a, tokenizer)
        tokens_b = tokenize(example.text_b, tokenizer)
        tokens_c = tokenize(example.text_c, tokenizer)

        # t_tokens = tokenize(example.entity, tokenizer)
        t_tokens = tokenizer(example.entity, add_special_tokens=False)["input_ids"]

        _truncate_seq_tuple(tokens_a, tokens_b, tokens_c, max_seq_length - 4)
        tokens_b = tokens_b + ["[SEP]"] + tokens_c

        tokens = []
        segment_ids = []
        tokens.append("[CLS]")
        segment_ids.append(0)
        for token in tokens_a:
            tokens.append(token)
            segment_ids.append(0)
        tokens.append("[SEP]")
        segment_ids.append(0)

        for token in tokens_b:
            tokens.append(token)
            segment_ids.append(1)
        tokens.append("[SEP]")
        segment_ids.append(1)

        input_ids = tokenizer.convert_tokens_to_ids(tokens)

        # The mask has 1 for real tokens and 0 for padding tokens. Only real
        # tokens are attended to.
        input_mask = [1] * len(input_ids)

        # Zero-pad up to the sequence length.
        while len(input_ids) < max_seq_length:
            input_ids.append(0)
            input_mask.append(0)
            segment_ids.append(0)


        assert len(input_ids) == max_seq_length
        assert len(input_mask) == max_seq_length
        assert len(segment_ids) == max_seq_length

        label_id = example.label 

        len_t = len(t_tokens)
        normal_input_ids = input_ids[:]
        for idx, input_id in enumerate(input_ids):
            if idx + len_t < len(input_ids) and input_ids[idx:idx+len_t] == t_tokens:
                # [MASK] id = 103
                for j in range(len_t):
                    input_ids[j+idx] = 103

        # append 1 sample with 2 input
        features.append(
            [InputFeatures(
                input_ids=normal_input_ids,
                input_mask=input_mask,
                segment_ids=segment_ids,
                label_id=label_id,
                entity = t_tokens
            ),
            InputFeatures(
                input_ids=input_ids,
                input_mask=input_mask,
                segment_ids=segment_ids,
                label_id=label_id,
                entity = t_tokens
            )]
        )
        
    print('#features', len(features))
    return features

def convert_examples_to_features_normal(examples, max_seq_length, tokenizer):
    print("#examples", len(examples))
    features = []
    for (ex_index, example) in enumerate(examples):
        tokens_a = tokenize(example.text_a, tokenizer)
        tokens_b = tokenize(example.text_b, tokenizer)
        tokens_c = tokenize(example.text_c, tokenizer)


        _truncate_seq_tuple(tokens_a, tokens_b, tokens_c, max_seq_length - 4)
        tokens_b = tokens_b + ["[SEP]"] + tokens_c
        
        
        inputs = tokenizer(
            example.text_a,
            example.text_b + tokenizer.sep_token + example.text_c,
            truncation="longest_first",
            max_length=max_seq_length,
            padding="max_length",
            add_special_tokens=True
        )

        # tokens = []
        # segment_ids = []
        # tokens.append("[CLS]")
        # segment_ids.append(0)
        # for token in tokens_a:
        #     tokens.append(token)
        #     segment_ids.append(0)
        # tokens.append("[SEP]")
        # segment_ids.append(0)

        # for token in tokens_b:
        #     tokens.append(token)
        #     segment_ids.append(1)
        # tokens.append("[SEP]")
        # segment_ids.append(1)

        # input_ids = tokenizer.convert_tokens_to_ids(tokens)

        # # The mask has 1 for real tokens and 0 for padding tokens. Only real
        # # tokens are attended to.
        # input_mask = [1] * len(input_ids)

        # # Zero-pad up to the sequence length.
        # while len(input_ids) < max_seq_length:
        #     input_ids.append(0)
        #     input_mask.append(0)
        #     segment_ids.append(0)

        # assert(inputs['input_ids'] == input_ids), print(inputs['input_ids'])

        # assert len(input_ids) == max_seq_length
        # assert len(input_mask) == max_seq_length
        # assert len(segment_ids) == max_seq_length

        label_id = example.label 

        if ex_index == 0:
            logger.info(f"input_text : {tokens_a} {tokens_b} {tokens_c}")
            logger.info(f"input_ids : {inputs['input_ids']}")
            logger.info(f"token_type_ids : {inputs['token_type_ids']}")
        
        # inputs = {}
        # inputs['input_ids'] = input_ids
        # inputs['attention_mask'] = input_mask
        # inputs['token_type_ids'] = segment_ids

        # append 1 sample with 2 input
        features.append(
            InputFeatures(
                input_ids=inputs['input_ids'],
                input_mask=inputs['attention_mask'],
                segment_ids=inputs['token_type_ids'],
                label_id=label_id,
            )
        )
        
    print('#features', len(features))
    return features



def convert_examples_to_features(examples, max_seq_length, tokenizer, args, rel2id):
    """Loads a data file into a list of `InputBatch`s."""

    save_file = "data/cached_wiki80.pkl"
    mode = "text"

    num_tokens = 0
    num_fit_examples = 0
    num_shown_examples = 0
    instances = []
    
    
    use_bert = "BertTokenizer" in tokenizer.__class__.__name__
    use_gpt = "GPT" in tokenizer.__class__.__name__
    
    assert not (use_bert and use_gpt), "model cannot be gpt and bert together"

    if False:
        with open(file=save_file, mode='rb') as fr:
            instances = pickle.load(fr)
        print('load preprocessed data from {}.'.format(save_file))

    else:
        print('loading..')
        for (ex_index, example) in enumerate(examples):
            

            """
                the relation between SUBJECT and OBJECT is .
                
            """

            if ex_index % 10000 == 0:
                logger.info("Writing example %d of %d" % (ex_index, len(examples)))

            tokens = []
            SUBJECT_START = "[subject_start]"
            SUBJECT_END = "[subject_end]"
            OBJECT_START = "[object_start]"
            OBJECT_END = "[object_end]"


            if mode.startswith("text"):
                for i, token in enumerate(example.sentence):
                    if i == example.span1[0]:
                        tokens.append(SUBJECT_START)
                    if i == example.span2[0]:
                        tokens.append(OBJECT_START)
                    # for sub_token in tokenizer.tokenize(token):
                    #     tokens.append(sub_token)
                    if i == example.span1[1]:
                        tokens.append(SUBJECT_END)
                    if i == example.span2[1]:
                        tokens.append(OBJECT_END)

                    tokens.append(token)

            SUBJECT = " ".join(example.sentence[example.span1[0]: example.span1[1]])
            OBJECT = " ".join(example.sentence[example.span2[0]: example.span2[1]])
            SUBJECT_ids = tokenizer(" "+SUBJECT, add_special_tokens=False)['input_ids']
            OBJECT_ids = tokenizer(" "+OBJECT, add_special_tokens=False)['input_ids']
            
            if use_gpt:
                if args.CT_CL:
                    prompt = f"[T1] [T2] [T3] [sub] {OBJECT} [sub] [T4] [obj] {SUBJECT} [obj] [T5] {tokenizer.cls_token}"
                else:
                    prompt = f"The relation between [sub] {SUBJECT} [sub] and [obj] {OBJECT} [obj] is {tokenizer.cls_token} ."
            else:
                # add prompt [T_n] and entity marker [obj] to enrich the context.
                prompt = f"[sub] {SUBJECT} [sub] {tokenizer.mask_token} [obj] {OBJECT} [obj] ."
            
            if ex_index == 0:
                input_text = " ".join(tokens)
                logger.info(f"input text : {input_text}")
                logger.info(f"prompt : {prompt}")
                logger.info(f"label : {example.label}")
            inputs = tokenizer(
                prompt,
                " ".join(tokens),
                truncation="longest_first",
                max_length=max_seq_length,
                padding="max_length",
                add_special_tokens=True
            )
            if use_gpt: cls_token_location = inputs['input_ids'].index(tokenizer.cls_token_id) 
            
            # find the subject and object tokens, choose the first ones
            sub_st = sub_ed = obj_st = obj_ed = -1
            for i in range(len(inputs['input_ids'])):
                if sub_st == -1 and inputs['input_ids'][i:i+len(SUBJECT_ids)] == SUBJECT_ids:
                    sub_st = i
                    sub_ed = i + len(SUBJECT_ids)
                if obj_st == -1 and inputs['input_ids'][i:i+len(OBJECT_ids)] == OBJECT_ids:
                    obj_st = i
                    obj_ed = i + len(OBJECT_ids)
            
            assert sub_st != -1 and obj_st != -1


            num_tokens += sum(inputs['attention_mask'])


            if sum(inputs['attention_mask']) > max_seq_length:
                pass
                # tokens = tokens[:max_seq_length]
            else:
                num_fit_examples += 1

            x = OrderedDict()
            x['input_ids'] = inputs['input_ids']
            if use_bert: x['token_type_ids'] = inputs['token_type_ids']
            x['attention_mask'] = inputs['attention_mask']
            x['label'] = rel2id[example.label]
            if use_gpt: x['cls_token_location'] = cls_token_location
            x['so'] =[sub_st, sub_ed, obj_st, obj_ed]

            instances.append(x)


        with open(file=save_file, mode='wb') as fw:
            pickle.dump(instances, fw)
        print('Finish save preprocessed data to {}.'.format( save_file))

    input_ids = [o['input_ids'] for o in instances]
    attention_mask = [o['attention_mask'] for o in instances]
    if use_bert: token_type_ids = [o['token_type_ids'] for o in instances]
    if use_gpt: cls_idx = [o['cls_token_location'] for o in instances]
    labels = [o['label'] for o in instances]
    so = torch.tensor([o['so'] for o in instances])


    input_ids = torch.tensor(input_ids)
    attention_mask = torch.tensor(attention_mask)
    if use_gpt: cls_idx = torch.tensor(cls_idx)
    if use_bert: token_type_ids = torch.tensor(token_type_ids)
    labels = torch.tensor(labels)

    logger.info("Average #tokens: %.2f" % (num_tokens * 1.0 / len(examples)))
    logger.info("%d (%.2f %%) examples can fit max_seq_length = %d" % (num_fit_examples,
                num_fit_examples * 100.0 / len(examples), max_seq_length))

    if use_gpt:
        dataset = TensorDataset(input_ids, attention_mask, cls_idx, labels)
    elif use_bert:
        dataset = TensorDataset(input_ids, attention_mask, token_type_ids, labels, so)
    else:
        dataset = TensorDataset(input_ids, attention_mask, labels)
    
    return dataset


def convert_examples_to_feature_sst2(examples, max_seq_length, tokenizer, args, rel2id):
    """Loads a data file into a list of `InputBatch`s."""

    save_file = "data/cached_wiki80.pkl"
    mode = "text"

    num_tokens = 0
    num_fit_examples = 0
    num_shown_examples = 0
    instances = []


    if False:
        with open(file=save_file, mode='rb') as fr:
            instances = pickle.load(fr)
        print('load preprocessed data from {}.'.format(save_file))

    else:
        print('loading..')
        for (ex_index, example) in enumerate(examples):
            try:
                prompt = f"[T1] [T2] {tokenizer.mask_token} ."
                inputs = tokenizer(
                    example.text_a + prompt,
                    truncation="longest_first",
                    max_length=max_seq_length,
                    padding="max_length",
                    add_special_tokens=True
                )

                x = OrderedDict()
                x['input_ids'] = inputs['input_ids']
                x['attention_mask'] = inputs['attention_mask']
                if "roberta" not in args.model_name_or_path:
                    x['token_type_ids'] = inputs['token_type_ids']
                x['label'] = int(example.label)


                instances.append(x)

            except Exception as e:
                print(e)

        with open(file=save_file, mode='wb') as fw:
            pickle.dump(instances, fw)
        print('Finish save preprocessed data to {}.'.format( save_file))

    input_ids = [o['input_ids'] for o in instances]
    attention_mask = [o['attention_mask'] for o in instances]

    if "roberta" not in args.model_name_or_path:
        token_type_ids = [o['token_type_ids'] for o in instances]
        token_type_ids = torch.tensor(token_type_ids)
    labels = [o['label'] for o in instances]


    input_ids = torch.tensor(input_ids)
    attention_mask = torch.tensor(attention_mask)
    labels = torch.tensor(labels)

    logger.info("Average #tokens: %.2f" % (num_tokens * 1.0 / len(examples)))
    logger.info("%d (%.2f %%) examples can fit max_seq_length = %d" % (num_fit_examples,
                num_fit_examples * 100.0 / len(examples), max_seq_length))
    if "roberta" not in args.model_name_or_path:
        dataset = TensorDataset(input_ids, attention_mask, token_type_ids, labels)
    else:
        dataset = TensorDataset(input_ids, attention_mask, labels)

    return dataset



def _truncate_seq_tuple(tokens_a, tokens_b, tokens_c, max_length):
    """Truncates a sequence tuple in place to the maximum length."""

    # This is a simple heuristic which will always truncate the longer sequence
    # one token at a time. This makes more sense than truncating an equal percent
    # of tokens from each, since if one sequence is very short then each token
    # that's truncated likely contains more information than a longer sequence.
    while True:
        total_length = len(tokens_a) + len(tokens_b) + len(tokens_c)
        if total_length <= max_length:
            break
        if len(tokens_a) >= len(tokens_b) and len(tokens_a) >= len(tokens_c):
            tokens_a.pop()
        elif len(tokens_b) >= len(tokens_a) and len(tokens_b) >= len(tokens_c):
            tokens_b.pop()
        else:
            tokens_c.pop()            

def get_dataset(mode, args, tokenizer, processor):

    if mode == "train":
        examples = processor.get_train_examples(args.data_dir)
    elif mode == "dev":
        examples = processor.get_dev_examples(args.data_dir)
    elif mode == "test":
        examples = processor.get_test_examples(args.data_dir)
    else:
        raise Exception("mode must be in choice [trian, dev, test]")
    gpt_mode = "wiki80" in args.task_name
  
    if "wiki80" in args.task_name:
        # normal relation extraction task
        dataset = convert_examples_to_features(
            examples, args.max_seq_length, tokenizer, args, processor.get_labels()
        )
        return dataset
    elif "sst" in args.task_name:
        dataset = convert_examples_to_feature_sst2(
            examples, args.max_seq_length, tokenizer, args, None
        )
        return dataset
    else:
        train_features = convert_examples_to_features_normal(
            examples, args.max_seq_length, tokenizer
        )


    input_ids = []
    input_mask = []
    segment_ids = []
    label_id = []
    entity_id = []

    for f in train_features:
        input_ids.append(f.input_ids)
        input_mask.append(f.input_mask)
        segment_ids.append(f.segment_ids)
        label_id.append(f.label_id)                

    all_input_ids = torch.tensor(input_ids, dtype=torch.long)
    all_input_mask = torch.tensor(input_mask, dtype=torch.long)
    all_segment_ids = torch.tensor(segment_ids, dtype=torch.long)
    all_label_ids = torch.tensor(label_id, dtype=torch.float)
   
    train_data = TensorDataset(all_input_ids, all_input_mask, all_segment_ids, all_label_ids)
    

    return train_data


def collate_fn(batch):
    pass




processors = {"normal": bertProcessor, "reloss": relossProcessor , "ptune": ptuneProcessor, "wiki80": wiki80Processor,
    "sst-2": Sst2Processor
}



# /usr/bin/env python
# coding=utf-8
"""utils"""
import logging
import os
import shutil
import json
from pathlib import Path

import torch

Label2IdxSub = {"B-H": 1, "I-H": 2, "O": 0}
Label2IdxObj = {"B-T": 1, "I-T": 2, "O": 0}


class Params:
    """参数定义
    """

    def __init__(self, ex_index=1, corpus_type='NYT'):
        # self.root_path = Path(os.path.abspath(os.path.dirname(__file__)))
        # self.data_dir = self.root_path / f'data/{corpus_type}'
        # self.ex_dir = self.root_path / f'experiments/ex{ex_index}'
        # self.model_dir = self.root_path / f'model/ex{ex_index}'
        # self.bert_model_dir = self.root_path / f'pretrain_models/bert_base_cased'

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.n_gpu = torch.cuda.device_count()
        # self.max_seq_length = 100
        # self.data_cache = False
        # self.train_batch_size = 6 if 'WebNLG' in corpus_type else 64
        # self.val_batch_size = 24
        # self.test_batch_size = 64
        # PRST parameters
        self.seq_tag_size = len(Label2IdxSub)
        # load label2id
        # self.rel2idx = json.load(open(self.data_dir/'rel2id.json', 'r', encoding='utf-8'))[-1]
        # self.rel_num = len(self.rel2idx)

        # early stop strategy
        # self.min_epoch_num = 20
        # self.patience = 0.00001
        # self.patience_num = 20

        # learning rate
        # self.fin_tuning_lr = 1e-4
        # self.downs_en_lr = 1e-3
        # self.clip_grad = 2.
        # self.drop_prob = 0.3  # dropout
        # self.weight_decay_rate = 0.01
        # self.warmup_prop = 0.1
        # self.gradient_accumulation_steps = 2

    def load(self, json_path):
        """Loads parameters from json file"""
        with open(json_path) as f:
            params = json.load(f)
            self.__dict__.update(params)

    def save(self, json_path):
        """保存配置到json文件
        """
        params = {}
        with open(json_path, 'w') as f:
            for k, v in self.__dict__.items():
                if isinstance(v, (str, int, float, bool)):
                    params[k] = v
            json.dump(params, f, indent=4)


class RunningAverage:
    """A simple class that maintains the running average of a quantity

    Example:
    ```
    loss_avg = RunningAverage()
    loss_avg.update(2)
    loss_avg.update(4)
    loss_avg() = 3
    ```
    """

    def __init__(self):
        self.steps = 0
        self.total = 0

    def update(self, val):
        self.total += val
        self.steps += 1

    def __call__(self):
        return self.total / float(self.steps)


def set_logger(save=False, log_path=None):
    """Set the logger to log info in terminal and file `log_path`.

    In general, it is useful to have a logger so that every output to the terminal is saved
    in a permanent file. Here we save it to `model_dir/train.log`.

    Example:
    ```
    logging.info("Starting training...")
    ```

    Args:
        log_path: (string) where to log
    """
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    if save and not os.path.exists(os.path.dirname(log_path)):
        os.makedirs(os.path.dirname(log_path))

    if not logger.handlers:
        if save:
            # Logging to a file
            file_handler = logging.FileHandler(log_path)
            file_handler.setFormatter(logging.Formatter('%(asctime)s:%(levelname)s: %(message)s'))
            logger.addHandler(file_handler)

        # Logging to console
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter('%(message)s'))
        logger.addHandler(stream_handler)


def save_checkpoint(state, is_best, checkpoint):
    """Saves model and training parameters at checkpoint + 'last.pth.tar'. If is_best==True, also saves
    checkpoint + 'best.pth.tar'

    Args:
        state: (dict) contains the entire model, may contain other keys such as epoch, optimizer
        is_best: (bool) True if it is the best model seen till now
        checkpoint: (string) folder where parameters are to be saved
    """
    filepath = os.path.join(checkpoint, 'last.pth.tar')
    if not os.path.exists(checkpoint):
        print("Checkpoint Directory does not exist! Making directory {}".format(checkpoint))
        os.makedirs(checkpoint)
    torch.save(state, filepath)
    if is_best:
        shutil.copyfile(filepath, os.path.join(checkpoint, 'best.pth.tar'))


def load_checkpoint(checkpoint, optimizer=True):
    """Loads entire model from file_path. If optimizer is True, loads
    optimizer assuming it is present in checkpoint.

    Args:
        checkpoint: (string) filename which needs to be loaded
        optimizer: (bool) resume optimizer from checkpoint
    """
    if not os.path.exists(checkpoint):
        raise ValueError("File doesn't exist {}".format(checkpoint))
    checkpoint = torch.load(checkpoint, map_location=torch.device('cpu'))

    if optimizer:
        return checkpoint['model'], checkpoint['optim']
    return checkpoint['model']



import torch
from torch import nn
from torchcrf import CRF
import torch.nn.functional as F
from .modeling_IFA import IFAModel
from transformers import BertConfig, BertModel, CLIPConfig, CLIPModel

class IFANERCRFModel(nn.Module):
    def __init__(self, label_list, args):
        super(IFANERCRFModel, self).__init__()
        self.args = args
        self.vision_config = CLIPConfig.from_pretrained(self.args.vit_name).vision_config
        self.text_config = BertConfig.from_pretrained(self.args.bert_name)

        clip_model_dict = CLIPModel.from_pretrained(self.args.vit_name).vision_model.state_dict()
        bert_model_dict = BertModel.from_pretrained(self.args.bert_name).state_dict()

        print(self.vision_config)
        print(self.text_config)

        self.vision_config.device = args.device
        self.model = IFAModel(self.vision_config, self.text_config)

        self.num_labels  = len(label_list) + 1  # pad
        self.crf = CRF(self.num_labels, batch_first=True)
        self.fc = nn.Linear(self.text_config.hidden_size, self.num_labels)
        self.dropout = nn.Dropout(0.1)

        # load:
        vision_names, text_names = [], []
        model_dict = self.model.state_dict()
        for name in model_dict:
            if 'vision' in name:
                clip_name = name.replace('vision_', '').replace('model.', '')
                if clip_name in clip_model_dict:
                    vision_names.append(clip_name)
                    model_dict[name] = clip_model_dict[clip_name]
            elif 'text' in name:
                text_name = name.replace('text_', '').replace('model.', '')
                if text_name in bert_model_dict:
                    text_names.append(text_name)
                    model_dict[name] = bert_model_dict[text_name]
        assert len(vision_names) == len(clip_model_dict) and len(text_names) == len(bert_model_dict), \
                    (len(vision_names), len(text_names), len(clip_model_dict), len(bert_model_dict))
        self.model.load_state_dict(model_dict)

    def forward(
            self, 
            input_ids=None, 
            attention_mask=None, 
            token_type_ids=None, 
            labels=None, 
            images=None, 
            aux_imgs=None,
            rcnn_imgs=None,
    ):
        bsz = input_ids.size(0)

        output = self.model(input_ids=input_ids,
                            attention_mask=attention_mask,
                            token_type_ids=token_type_ids,

                            pixel_values=images,
                            aux_values=aux_imgs, 
                            rcnn_values=rcnn_imgs,
                            return_dict=True,)

        sequence_output = output.last_hidden_state       # bsz, len, hidden
        sequence_output = self.dropout(sequence_output)  # bsz, len, hidden
        emissions = self.fc(sequence_output)             # bsz, len, labels
        
        logits = self.crf.decode(emissions, attention_mask.byte())
        loss = None
        if labels is not None:
            loss = -1 * self.crf(emissions, labels, mask=attention_mask.byte(), reduction='mean')  # 去掉CLS
            return logits, loss
        return logits, None


"""
    Autoregressive Structured Prediction T5 model for Named Entity Recognition
    Tianyu Liu
"""
import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../")))
from ctypes import sizeof
import math
from typing import Optional, Tuple, Any, Dict, Iterable, List
import logging

import torch
import torch.utils.checkpoint
from torch import nn
import torch.nn.init as init
import torch.nn.functional as F

from transformers.models.t5.modeling_t5 import (
    T5PreTrainedModel, T5Model,
    T5_INPUTS_DOCSTRING, _CONFIG_FOR_DOC
)
from transformers.file_utils import (
    add_start_docstrings_to_model_forward,
    replace_return_docstrings
)

from .modeling_outputs import MySeq2SeqLMOutput

import util
from util import (
    batched_masked_select,
    logsumexp
)
logger = logging.getLogger()
NEGINF = -20000.

class T5NER(T5PreTrainedModel):
    base_model_prefix = "t5"

    def __init__(
        self,
        config,
        asp_hidden_dim: int = 1500,
        asp_dropout_rate: float = 0.3,
        asp_init_std: float = 0.02,
        asp_feature_emb_size: int = 20,
        asp_linking_distance_num_buckets: int = 16,
        asp_activation: str='relu',
        num_typing_classes: int = 4,
        mention_start_id: int = 0,
        mention_end_id: int = 0,
    ):
        super().__init__(config)
        self.t5 = T5Model.from_pretrained(self.config._name_or_path)
        self.dropout = nn.Dropout(p=asp_dropout_rate)
        self.init_std = asp_init_std

        self.feature_emb_size = asp_feature_emb_size
        self.linking_distance_num_buckets = asp_linking_distance_num_buckets
        # output_dim: {0, 1, 2}
        # corresponding to 3 actions: {mention_end, mention_start, copy}

        self.num_typing_classes = num_typing_classes
        # conll04, ace04, ace05 has no nested mention
        self.max_nest_depth = 1
        self.action_head = util.make_ffnn(
            self.config.d_model,
            hidden_size=[asp_hidden_dim],
            output_size=1,
            dropout=self.dropout,
            std=self.init_std,
            activation=asp_activation
        )

        lr_scorer_input_dim = 2*self.config.d_model
        # left-right bracket
        self.lr_scorer = util.make_ffnn(
            lr_scorer_input_dim,
            hidden_size=[asp_hidden_dim],
            output_size=self.num_typing_classes,
            dropout=self.dropout,
            std=self.init_std,
            activation=asp_activation
        )
        self.last_device = torch.cuda.device_count() - 1

        self.mention_start_id = mention_start_id
        self.mention_end_id = mention_end_id
        self.post_init()


    @add_start_docstrings_to_model_forward(T5_INPUTS_DOCSTRING)
    @replace_return_docstrings(output_type=MySeq2SeqLMOutput, config_class=_CONFIG_FOR_DOC)
    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        decoder_input_ids=None,
        decoder_attention_mask=None,
        head_mask=None,
        decoder_head_mask=None,
        cross_attn_head_mask=None,
        encoder_outputs=None,
        past_key_values=None,
        inputs_embeds=None,
        decoder_inputs_embeds=None,
        labels=None,
        sentence_map=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        encoder_input_ids=None,
        full_decoder_input_ids=None,
        full_hidden_states=None,
        lr_pair_flag=None,
        sentence_idx=None,
        decoder_pairing=None,
        decoder_typing=None,
    ):
        r"""
        Inputs:
            labels (`torch.LongTensor` of shape `(batch_size, target_sequence_length)`, *optional*, *training*):
                Labels for computing the masked language modeling loss. 
                Indices should either be in `[0, 1, 2]` (3 basic actions) or -1 (see `input_ids` docstring). 
                labels with <0 values are ignored (masked).
            lr_pair_flag (`torch.LongTensor` of shape `(batch_size, sequence_length, number_of_l_brackets)`, *optional*, *training*):
                A flag tensor. Indicating whether a pair of tokens are paired *left-right* brackts. 1 for paired, 0 for not.
            rr_pair_flag (`torch.LongTensor` of shape `(batch_size, sequence_length, number_of_r_brackets)`, *optional*, *training*):
                A flag tensor. Indicating whether a pair of tokens are paired *right-right* brackts. 1 for paired, 0 for not.
            full_decoder_input_ids (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*, *inference*):
                Decoded token ids until now. Keeping track of how many tokens have been copied from
                the input, so we know which token to copy now.
            full_hidden_states (`torch.FloatTensor` of shape `(batch_size, sequence_length, hidden_size)`, *optional*, *inference*):
                Hidden states of the decoded tokens until now. 
            decoder_pairing (`List[torch.LongTensor]` each element is shape `(batch_size, 1)`, *optional*, *inference*):
                A list of tensors. Each tensor is a tensor of shape `(batch_size, 1)`, decoded pairing actions until now.
            decoder_typing (`List[torch.LongTensor]` each element is shape `(batch_size, 1)`, *optional*, *inference*):
                A list of tensors. Each tensor is a tensor of shape `(batch_size, 1)`, decoded typing actions until now.
        Returns:
            output (`MySeq2SeqLMOutput` ):
                loss (`torch.FloatTensor` of shape `(1)`, *optional*, returned when `labels` is provided):
                pairing (`torch.LongTensor` of shape `(batch_size, 1)`, pairing actions of this step, *inferece* only)
                typing (`torch.LongTensor` of shape `(batch_size, 1)`, typing actions of this step, *inferece* only)
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict
        target_ids = decoder_input_ids
        if labels is not None:  # training
            # decoder_input_ids starts with <pad> and has the same length as target_ids
            decoder_input_ids = self._shift_right(
                target_ids
            )

        outputs = self.t5(
            input_ids,
            attention_mask=attention_mask,
            decoder_input_ids=decoder_input_ids,
            encoder_outputs=encoder_outputs,
            decoder_attention_mask=decoder_attention_mask,
            head_mask=head_mask,
            decoder_head_mask=decoder_head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            decoder_inputs_embeds=decoder_inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        loss, lm_logits = None, None
        if labels is not None:  # Training: We compute loss
            outputs.last_hidden_state = outputs.last_hidden_state.to(self.device)

            # shape: (batchsize, seq_len, 1)
            action_logits = self.action_head(outputs.last_hidden_state)

            # a new loss that is compatible with inference
            (numer, denom) = self.get_logits_training(
                outputs.last_hidden_state,  # current_hidden_state
                target_ids=target_ids, 
                lr_pair_flag=lr_pair_flag,
                labels=labels
            )
            # keeping <copy> score 0.
            action_logits = torch.cat(
                [torch.zeros_like(action_logits),
                 action_logits], dim=-1
            )
            # Note: We want to compute a joint log-likelihood for action, boundary, and antecedent,
            # And we will use this joint LL for inference with beam search
            denom = util.logsumexp(torch.cat([
                action_logits,
                denom
            ], dim=-1), dim=-1)
            numer = util.logsumexp(torch.cat([
                action_logits + torch.where(
                    util.one_hot_ignore_negative(
                        labels, num_classes=3), 0., float("-inf")
                )[..., :2],
                numer
            ], dim=-1), dim=-1)

            loss = (denom - numer)[decoder_attention_mask.bool()].sum()
            loss = loss / target_ids.size(0)

            l_choice, typing_choice = None, None

        else:  # inference, step-by-step classifications
            # (batch_size, 1, dim)
            outputs.last_hidden_state = outputs.last_hidden_state.to(self.device)
            # denom: (batch_size, 1, 1)
            # l_choice / typing_choice: (batch_size, 1)
            (denom, l_choice, typing_choice) = self.get_logits_inference(
                outputs.last_hidden_state,  # current_hidden_state
                full_decoder_input_ids,     # previous decoded ids
                full_hidden_states=full_hidden_states
            )
            action_logits = self.action_head(outputs.last_hidden_state)
            # (batch_size, 1, 1)
            action_logits = torch.cat(
                [torch.zeros_like(action_logits),
                 action_logits,
                 denom],
                dim=-1
            )
            # Restore lm_logits from action_logits
            lm_logits = self.decoder_input_ids_to_vocab_mask(
                action_logits,
                full_decoder_input_ids,
                encoder_input_ids
            )

        return MySeq2SeqLMOutput(
            loss=loss,
            logits=lm_logits,
            past_key_values=outputs.past_key_values,
            decoder_hidden_states=outputs.decoder_hidden_states,
            decoder_attentions=outputs.decoder_attentions,
            cross_attentions=outputs.cross_attentions,
            encoder_last_hidden_state=outputs.encoder_last_hidden_state,
            encoder_hidden_states=outputs.encoder_hidden_states,
            encoder_attentions=outputs.encoder_attentions,
            pairing=l_choice,
            typing=typing_choice
        )

    def get_logits_inference(
        self,
        current_hidden_state,  # inference: (batch_size, 1, dim)
        decoder_input_ids,  # (batch_size, seq_len, )
        full_hidden_states=None,  # tuple((batch_size, 1, dim))
    ):
        # Shape: (batch_size, seq_len, )
        batch_size = decoder_input_ids.size(0)
        output_ids = decoder_input_ids[:, 1:]  # excluding decoder BOS

        range_vec = torch.arange(
            output_ids.size(1)-1, -1, -1,
            dtype=torch.long, device=self.device
        ).unsqueeze(0).expand(batch_size, -1)

        if len(full_hidden_states) == 0:
            # the first valid token in the output
            return (
                current_hidden_state.new_full((batch_size, 1, self.num_typing_classes), float("-inf")), 
                decoder_input_ids.new_full((batch_size, 1), -1),
                decoder_input_ids.new_full((batch_size, 1), -1)
            )

        # concatenating into sequence: (batch_size, seq_len, dim)
        decoder_output = torch.cat(full_hidden_states, dim=1)

        # check special tokens
        # Shape: (batch_size, seq_len, )
        is_l = (output_ids == self.mention_start_id)

        if is_l.sum() == 0:
            # no full mention and no previous mentions
            lr_denom = current_hidden_state.new_full((batch_size, 1, self.num_typing_classes), float("-inf"))
            l_choice = decoder_input_ids.new_full((batch_size, 1), -1)
            typing_choice = decoder_input_ids.new_full((batch_size, 1), -1)
            denom = lr_denom
        else:
            # (batch_size, num_l, dim), (batch_size, num_l, )
            l_emb, l_emb_mask = batched_masked_select(decoder_output, is_l)
            
            # (batch_size, 1, num_l, )
            distance_to_previous_l, _ = batched_masked_select(range_vec, is_l)
            distance_to_previous_l = distance_to_previous_l.unsqueeze(1)

            # (batch_size, 1, num_l, 2*dim+feature_dim)
            lr_pair_emb = util.prepare_pair_embeddings(
                l_emb, current_hidden_state)

            # (batch_size, 1, num_l, self.num_typing_classes)
            lr_score = self.lr_scorer(lr_pair_emb)
            
            num_l_each_instance = is_l.sum(dim=-1)

            for i in range(batch_size):
                lr_score[i, :, :num_l_each_instance[i]-self.max_nest_depth, :] = NEGINF
                lr_score[i, :, num_l_each_instance[i]:, :] = NEGINF
                
            # (batch_size, 1)
            lr_denom = util.logsumexp(
                lr_score, dim=(2, 3)
            ).unsqueeze(-1)

            # (batch_size, 1, self.num_typing_classes)
            lr_score_max_over_entities, max_l = lr_score.max(dim=2)
            # (batch_size, 1)
            typing_choice = lr_score_max_over_entities.argmax(dim=2)
            # (batch_size, 1)
            l_choice = max_l.squeeze(1).gather(1, typing_choice)
            denom = lr_denom

        return (denom, l_choice, typing_choice)


    def get_logits_training(
        self,
        decoder_output,  # (batch_size, seq_len, hidden_dim)
        target_ids,      # (batch_size, seq_len, )
        lr_pair_flag=None,
        labels=None
    ):
        batch_size, seq_len = decoder_output.size(0), decoder_output.size(1)
        target_mask = (target_ids != self.config.pad_token_id)
        # (batch_size, seq_len)
        linearized_indices = (labels >= 0).cumsum(dim=-1) - 1
        # (batch_size, seq_len)
        is_l = (target_ids == self.mention_start_id)
        is_r = (target_ids == self.mention_end_id)

        if is_r.sum() == 0:
            numer, denom = (torch.full_like(decoder_output[...,:1], NEGINF),
                            torch.full_like(decoder_output[...,:1], NEGINF))
            return numer, denom

        # (batch_size, num_r / num_l)
        (l_pos, l_pos_mask) = batched_masked_select(linearized_indices, is_l)

        # (batch_size, num_r / num_l, hidden_dim)
        l_emb, _ = batched_masked_select(decoder_output, is_l)
        
        # (batch_size, seq_len, num_r / num_l)
        distance_to_previous_l = linearized_indices.unsqueeze(2) - l_pos.unsqueeze(1)
        
        # (batch_size, seq_len, num_r / num_l)
        is_after_l = (distance_to_previous_l > 0)
        is_after_l = is_after_l & target_mask.unsqueeze(2) & l_pos_mask.unsqueeze(1)

        # check correctness for batch
        kept_l = min(self.max_nest_depth, l_emb.size(1))
        # (batch_size, seq_len, kept_l)
        _, prev_l_indices = (
            -distance_to_previous_l + (is_after_l * 10000)
        ).topk(kept_l, dim=2)

        # (batch_size, seq_len, kept_l, hidden_dim)
        kept_l_emb = util.batch_select(l_emb, prev_l_indices)
        # (batch_size, seq_len, kept_l)
        distance_to_previous_l = util._batched_index_select(
            distance_to_previous_l, prev_l_indices, dim=2
        )

        expanded_decoder_output = decoder_output.unsqueeze(2).expand(
            -1, -1, kept_l, -1
        )
        # shape(batch_size, seq_len, kept_l, 2*hidden_dim)
        lr_pair_emb = torch.cat([kept_l_emb, expanded_decoder_output], dim=-1)

        # shape(batch_size, seq_len, kept_l, num_typing_classes)
        kept_is_after_l = is_after_l.gather(dim=2, index=prev_l_indices)
        lr_score = self.lr_scorer(lr_pair_emb) + (~kept_is_after_l).unsqueeze(-1) * NEGINF

        # (batch_size, seq_len, 1)
        lr_denom = logsumexp(
            lr_score,
            dim=(2, 3), keepdim=False
        ).unsqueeze(-1) * is_after_l.any(dim=2, keepdim=True)
        # (batch_size, seq_len, num_l, num_typing_classes) ->
        #     (batch_size, seq_len, kept_l, num_typing_classes)
        kept_lr_pair_flag = util._batched_index_select(
            lr_pair_flag, prev_l_indices, dim=2
        )
        # (batch_size, seq_len, 1)
        lr_numer = logsumexp(
            lr_score + (~kept_lr_pair_flag) * NEGINF,
            dim=(2, 3), keepdim=False
        ).unsqueeze(-1) * is_after_l.any(dim=2, keepdim=True)

        numer, denom = lr_numer, lr_denom

        return numer, denom


    def decoder_input_ids_to_vocab_mask(
        self,
        action_logits,
        decoder_input_ids,
        input_ids
    ):
        # counting how many words have been copied
        is_copied = ((decoder_input_ids != self.mention_start_id) &\
                     (decoder_input_ids != self.mention_end_id))
        # (batch_size, 1)
        num_copied = is_copied.sum(dim=-1, keepdim=True) - 1
        num_copied = num_copied.clamp(max=input_ids.size(1) - 1)

        # compute pointer to input tokens
        lm_logits = action_logits.new_full(
            (action_logits.size(0), action_logits.size(1), self.config.vocab_size),
            float("-inf")
        )
        word_to_copy = input_ids.expand(num_copied.size(0), -1)  # repeating over beams
        word_to_copy = word_to_copy.gather(1, num_copied)

        lm_logits.scatter_(2, word_to_copy.unsqueeze(-1), action_logits[:, :, :1])
        lm_logits[:, :, self.mention_start_id] = action_logits[:, :, 1]
        lm_logits[:, :, self.mention_end_id] = action_logits[:, :, 2]

        return lm_logits

    def prepare_inputs_for_generation(
        self,
        decoder_input_ids,
        past=None,
        attention_mask=None,
        head_mask=None,
        decoder_head_mask=None,
        cross_attn_head_mask=None,
        use_cache=None,
        encoder_outputs=None,
        decoder_encoder_input_ids=None,
        **kwargs
    ):
        # cut decoder_input_ids if past is used
        if past is not None:
            cut_decoder_input_ids = decoder_input_ids[:, -1:]
        else:
            cut_decoder_input_ids = decoder_input_ids

        if "full_hidden_states" not in kwargs:
            kwargs["full_hidden_states"] = []  # initializing the list

        return {
            "input_ids": None,  # encoder_outputs is defined. input_ids not needed
            "encoder_input_ids": decoder_encoder_input_ids,
            "encoder_outputs": encoder_outputs,
            "past_key_values": past,
            "decoder_input_ids": cut_decoder_input_ids,   # last decoder_input_ids
            "full_decoder_input_ids": decoder_input_ids,  # full_decoder_input_ids
            "full_hidden_states": kwargs["full_hidden_states"],
            "decoder_pairing": kwargs["decoder_pairing"],
            "decoder_typing": kwargs["decoder_typing"],
            "attention_mask": attention_mask,
            "head_mask": head_mask,
            "decoder_head_mask": decoder_head_mask,
            "cross_attn_head_mask": cross_attn_head_mask,
            "use_cache": use_cache,
        }

    def prepare_decoder_input_ids_from_labels(self, labels: torch.Tensor):
        return self._shift_right(labels)

    # override
    @staticmethod
    def _update_model_kwargs_for_generation(
        outputs: MySeq2SeqLMOutput, model_kwargs: Dict[str, Any], is_encoder_decoder: bool = False
    ) -> Dict[str, Any]:
        # update past
        if "past_key_values" in outputs:
            model_kwargs["past"] = outputs.past_key_values
        elif "mems" in outputs:
            model_kwargs["past"] = outputs.mems
        elif "past_buckets_states" in outputs:
            model_kwargs["past"] = outputs.past_buckets_states
        else:
            model_kwargs["past"] = None

        if "full_hidden_states" not in model_kwargs:
            model_kwargs["full_hidden_states"] = []

        model_kwargs["full_hidden_states"].append(
            outputs.decoder_hidden_states[-1].to(outputs.pairing.get_device())
        )
        model_kwargs["decoder_pairing"].append(outputs.pairing)
        model_kwargs["decoder_typing"].append(outputs.typing)

        # update token_type_ids with last value
        if "token_type_ids" in model_kwargs:
            token_type_ids = model_kwargs["token_type_ids"]
            model_kwargs["token_type_ids"] = torch.cat(
                [token_type_ids, token_type_ids[:, -1].unsqueeze(-1)], dim=-1)

        # update attention mask
        if not is_encoder_decoder:
            if "attention_mask" in model_kwargs:
                attention_mask = model_kwargs["attention_mask"]
                model_kwargs["attention_mask"] = torch.cat(
                    [attention_mask, attention_mask.new_ones((attention_mask.size(0), 1))], dim=-1
                )
        return model_kwargs

    def get_encoder(self):
        return self.t5.get_encoder()

    def get_decoder(self):
        return self.t5.get_decoder()

    def resize_token_embeddings(self, new_num_tokens: int) -> nn.Embedding:
        new_embeddings = super().resize_token_embeddings(new_num_tokens)
        return new_embeddings

    def _resize_action_logits_bias(self, new_num_tokens: int) -> None:
        old_num_tokens = self.action_logits_bias.shape[-1]
        if new_num_tokens <= old_num_tokens:
            new_bias = self.action_logits_bias[:, :new_num_tokens]
        else:
            extra_bias = torch.zeros(
                (1, new_num_tokens - old_num_tokens), device=self.device)
            new_bias = torch.cat([self.action_logits_bias, extra_bias], dim=1)
        self.register_buffer("action_logits_bias", new_bias)

    def get_output_embeddings(self):
        return None

    def get_action_head(self):
        return self.action_head

    def set_output_embeddings(self, new_embeddings):
        self.action_head = new_embeddings

    @staticmethod
    def _reorder_cache(past, beam_idx):
        reordered_past = ()
        for layer_past in past:
            # cached cross_attention states don't have to be reordered -> they are always the same
            reordered_past += (
                tuple(past_state.index_select(0, beam_idx)
                      for past_state in layer_past[:2]) + layer_past[2:],
            )
        return reordered_past



"""
    Sequence to sequence model wrapper for coreference resolution.
    Decoding algorithms, parallelism handling, and other utilities
    Tianyu Liu
"""
import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../")))
import torch
import torch.nn as nn
import logging

from transformers import T5Tokenizer
from .t5_coref import T5Coref

logger = logging.getLogger(__file__)


class CorefModel(torch.nn.Module):
    """
        Model wrapper for coreference resolution.
    """
    def __init__(self, config, device):
        super().__init__()
        self.config = config
        self.max_seg_len = config["max_segment_len"]
        self.device = device # main device

        # Model
        self.dropout = nn.Dropout(p=config["dropout_rate"])

        if 'device_map' in config:
            self.config['device_map'] = {
                int(k): v for k, v in config['device_map'].items()
            }

        self.tz = T5Tokenizer.from_pretrained("t5-small")

        self.SPEAKER_START = '<speaker>'
        self.SPEAKER_END   = '</speaker>'
        self.MENTION_START = '<m>'
        self.MENTION_END   = '</m>'

        self.tz.add_tokens(self.SPEAKER_START)
        self.tz.add_tokens(self.SPEAKER_END)
        self.tz.add_tokens(self.MENTION_START)
        self.tz.add_tokens(self.MENTION_END)

        self.mention_start_id = self.tz.convert_tokens_to_ids(self.MENTION_START)
        self.mention_end_id   = self.tz.convert_tokens_to_ids(self.MENTION_END)

        self.model = T5Coref.from_pretrained(
            config["plm_pretrained_name_or_path"],
            asp_hidden_dim=config["hidden_size"],
            asp_dropout_rate=config["dropout_rate"],
            asp_init_std=config["init_std"],
            asp_feature_emb_size=config["feature_emb_size"],
            asp_activation=config["activation"],
            mention_start_id=self.mention_start_id,
            mention_end_id=self.mention_end_id
        )

        self.beam_size = config["beam_size"]
        self.model.resize_token_embeddings(self.tz.vocab_size + 4)


    def parallel_preparation_training(self, ):
        if torch.cuda.device_count() == 1:
            self.model = self.model.cuda()
            return
        # prepare the model for parallel training
        if (not self.model.t5.model_parallel or
                self.model.emb_rr_distance.weight.get_device() != self.device):
            logger.info(
                f"Moving model to {self.device} and parallelize for training"
            )
            if not self.model.t5.model_parallel:
                self.model.t5.parallelize(
                    device_map=self.config['device_map'] if 'device_map' in self.config else None)
            self.model.rr_scorer = self.model.rr_scorer.to(self.device)
            self.model.lr_scorer = self.model.lr_scorer.to(self.device)
            self.model.action_head = self.model.action_head.to(self.device)
            self.model.emb_rr_distance = self.model.emb_rr_distance.to(self.device)
            torch.cuda.empty_cache()
        return


    def parallel_preparation_inference(self, ):
        if torch.cuda.device_count() == 1:
            self.model = self.model.cuda()
            return
        # prepare the model for parallel inference
        if self.model.emb_rr_distance.weight.get_device() != self.device:
            logger.info(
                f"Moving model from {self.model.emb_rr_distance.weight.get_device()} to {self.device} for inference"
            )
            self.model.rr_scorer = self.model.rr_scorer.to(self.device)
            self.model.lr_scorer = self.model.lr_scorer.to(self.device)
            self.model.action_head = self.model.action_head.to(self.device)
            self.model.emb_rr_distance = self.model.emb_rr_distance.to(self.device)
            if not self.model.t5.model_parallel:
                self.model.t5.parallelize()
            torch.cuda.empty_cache()
        return


    def get_params(self, named=False):
        plm_based_param, task_param = [], []
        for name, param in self.named_parameters():
            if 't5' in name:
                to_add = (name, param) if named else param
                plm_based_param.append(to_add)
            else:
                to_add = (name, param) if named else param
                task_param.append(to_add)
        return plm_based_param, task_param


    def forward(
        self,
        input_ids, input_mask,
        target_ids, target_mask,
        action_labels, lr_pair_flag, rr_pair_flag,
        is_training,
        **kwargs,
    ):
        if len(is_training.size()) == 1:
            is_training = is_training[0]

        if (is_training == 1):  # training
            self.parallel_preparation_training()

            flag_grad_ckpt = False
            # reduce the memory usage with gradient checkpointing
            if target_ids.size(1) > 1400 and\
                ("3b" in self.config['plm_pretrained_name_or_path'].lower() or
                 "-xl" in self.config['plm_pretrained_name_or_path'].lower() or
                 "-xxl" in self.config['plm_pretrained_name_or_path'].lower()):
                self.model.gradient_checkpointing_enable()
                flag_grad_ckpt = True
            seq2seq_output = self.model(
                input_ids=input_ids,
                attention_mask=input_mask,
                decoder_input_ids=target_ids,
                decoder_attention_mask=target_mask,
                labels=action_labels,
                output_hidden_states=True,
                lr_pair_flag=lr_pair_flag,
                rr_pair_flag=rr_pair_flag,
                use_cache=(not flag_grad_ckpt)
            )
            if flag_grad_ckpt:
                self.model.gradient_checkpointing_disable()
                flag_grad_ckpt = False
            total_loss = seq2seq_output.loss

            return total_loss

        else:  # inference
            self.parallel_preparation_inference()

            # save the decoded actions
            decoder_pairing, decoder_linking = [], []
            model_output = self.model.generate(
                input_ids,
                early_stopping=True,
                max_length=4096,
                num_beams=self.beam_size,
                num_return_sequences=self.beam_size,
                no_repeat_ngram_size=0,
                encoder_no_repeat_ngram_size=0,
                return_dict_in_generate=True,
                output_hidden_states=True,
                output_scores=True,
                **{
                    "decoder_encoder_input_ids": input_ids,
                    "decoder_pairing": decoder_pairing,
                    "decoder_linking": decoder_linking
                }
            )
            results = {
                "output_ids": [],
                "start_token": [],
                "end_token": [],
                "pairing": [],
                "linking": []
            }

            for i in range(input_ids.size(0)):
                # Separatly decoding each instance in the batch
                # removing </s>
                output_ids = model_output.sequences[i][1:]

                # check special tokens
                is_start_token = (output_ids == self.mention_start_id)
                is_end_token = (output_ids == self.mention_end_id)

                range_vec = torch.arange(0, output_ids.size(0), device=self.device)
                start_token_pos = range_vec[is_start_token]
                end_token_pos = range_vec[is_end_token]
                results["output_ids"].append(output_ids)
                results["start_token"].append(start_token_pos)
                results["end_token"].append(end_token_pos)

                results["pairing"].append([x[i] for x in decoder_pairing])
                results["linking"].append([x[i] for x in decoder_linking])

            return results


    def get_mapping_to_input_sequence(
        self, output_ids
    ):
        # Get the mapping from the output with special tokens
        # to the input without special tokens.
        mapping, new_id = [], -1
        for i in range(len(output_ids)):
            if output_ids[i] == self.mention_start_id:
                new_id += 1
            elif output_ids[i] == self.mention_end_id:
                new_id += 0
            else:
                new_id += 1
            mapping.append(new_id)
            if output_ids[i] == self.mention_start_id:
                new_id -= 1

        return mapping


    def decoding(self, output, stored_info):
        output_ids, pairing_decisions, antecedent_decisions = (
            output["output_ids"].tolist(),
            output["pairing"],
            output["linking"]
        )
        subtoken_map = stored_info["subtoken_map"]
        # mapping from generated sequence index to original index
        mapping = self.get_mapping_to_input_sequence(output_ids)

        mentions, start_ind = [], []
        predicted_clusters, antecedent_indices = [], []

        mention_to_cluster_id = {}

        for i in range(len(output_ids)):
            if output_ids[i] == self.tz.pad_token_id:
                break
            if output_ids[i] == self.mention_start_id:
                start_ind.append(i)
            if output_ids[i] == self.mention_end_id:
                mention = (
                    subtoken_map[mapping[start_ind[pairing_decisions[i]]]],
                    subtoken_map[mapping[i]]
                )
                mentions.append(mention)
                ant = antecedent_decisions[i]
                if ant == -1:
                    predicted_clusters.append([mention])
                    mention_to_cluster_id[mention] = len(predicted_clusters) - 1
                else:
                    antecedent = mentions[ant]
                    predicted_clusters[
                        mention_to_cluster_id[antecedent]
                    ].append(mention)
                    mention_to_cluster_id[mention] = mention_to_cluster_id[antecedent]

        mention_to_predicted = {
            m: tuple(predicted_clusters[cluster_idx])\
            for m, cluster_idx in mention_to_cluster_id.items()
        }
        predicted_clusters = [tuple(c) for c in predicted_clusters]

        result_dict = {
            "predicted": predicted_clusters,
            "mention_to_predicted": mention_to_predicted,
            "predicted_mentions": mentions
        }
        return result_dict

    def extract_gold_clusters_from_gold_annotation(
        self, 
        stored_info
    ):
        output_ids = stored_info['target_sentence']
        cluster_category = stored_info['cluster_category']
        mention_indice = stored_info['mention_indice']

        if type(output_ids[0]) == str:
            output_ids = [self.tz.convert_tokens_to_ids(x) for x in output_ids]
        # mapping from generated sequence index to original index
        mapping = self.get_mapping_to_input_sequence(output_ids)
        subtoken_map = stored_info["subtoken_map"]

        # Get predicted clusters
        mention_to_cluster_id = {}
        gold_clusters, mentions = [], []

        for i, (category_idx, boundary_idx) in enumerate(zip(cluster_category, mention_indice)):
            if category_idx == -1:
                continue
            # Add mention to cluster
            mention = (
                int(subtoken_map[mapping[boundary_idx]]), 
                int(subtoken_map[mapping[i]])
            )
            mentions.append(mention)
            while category_idx >= len(gold_clusters):
                gold_clusters.append([])

            gold_clusters[category_idx].append(mention)
            mention_to_cluster_id[mention] = category_idx

        gold_clusters = [tuple(c) for c in gold_clusters if len(c) > 1]
        mention_to_gold = {m: cluster for cluster in gold_clusters for m in cluster}

        result_dict = {
            "gold": gold_clusters,
            "mention_to_gold": mention_to_gold,
            "gold_mentions": mentions
        }
        return result_dict



import logging
import pickle
import time
from collections import defaultdict, deque


def get_logger(dataset):
    pathname = "./log/{}_{}.txt".format(dataset, time.strftime("%m-%d_%H-%M-%S"))
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter("%(asctime)s - %(levelname)s: %(message)s",
                                  datefmt='%Y-%m-%d %H:%M:%S')

    file_handler = logging.FileHandler(pathname)
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(formatter)

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.DEBUG)
    stream_handler.setFormatter(formatter)

    logger.addHandler(file_handler)
    logger.addHandler(stream_handler)

    return logger


def save_file(path, data):
    with open(path, "wb") as f:
        pickle.dump(data, f)


def load_file(path):
    with open(path, "rb") as f:
        data = pickle.load(f)
    return data


def convert_index_to_text(index, type):
    text = "-".join([str(i) for i in index])
    text = text + "-#-{}".format(type)
    return text


def convert_text_to_index(text):
    index, type = text.split("-#-")
    index = [int(x) for x in index.split("-")]
    return index, int(type)


def decode(outputs, entities, length):
    class Node:
        def __init__(self):
            self.THW = []  # [(tail, type)]
            self.NNW = defaultdict(set)  # {(head,tail): {next_index}}

    ent_r, ent_p, ent_c = 0, 0, 0
    decode_entities = []
    q = deque()
    for instance, ent_set, l in zip(outputs, entities, length):
        predicts = []
        nodes = [Node() for _ in range(l)]
        for cur in reversed(range(l)):
            heads = []
            for pre in range(cur + 1):
                # THW
                if instance[cur, pre] > 1:
                    nodes[pre].THW.append((cur, instance[cur, pre]))
                    heads.append(pre)
                # NNW
                if pre < cur and instance[pre, cur] == 1:
                    # cur node
                    for head in heads:
                        nodes[pre].NNW[(head, cur)].add(cur)
                    # post nodes
                    for head, tail in nodes[cur].NNW.keys():
                        if tail >= cur and head <= pre:
                            nodes[pre].NNW[(head, tail)].add(cur)
            # entity
            for tail, type_id in nodes[cur].THW:
                if cur == tail:
                    predicts.append(([cur], type_id))
                    continue
                q.clear()
                q.append([cur])
                while len(q) > 0:
                    chains = q.pop()
                    for idx in nodes[chains[-1]].NNW[(cur, tail)]:
                        if idx == tail:
                            predicts.append((chains + [idx], type_id))
                        else:
                            q.append(chains + [idx])

        predicts = set([convert_index_to_text(x[0], x[1]) for x in predicts])
        decode_entities.append([convert_text_to_index(x) for x in predicts])
        ent_r += len(ent_set)
        ent_p += len(predicts)
        ent_c += len(predicts.intersection(ent_set))
    return ent_c, ent_p, ent_r, decode_entities


def cal_f1(c, p, r):
    if r == 0 or p == 0:
        return 0, 0, 0

    r = c / r if r else 0
    p = c / p if p else 0

    if r and p:
        return 2 * p * r / (p + r), p, r
    return 0, p, r


#!/usr/bin/env python
# -*- coding:utf-8 -*-
from dataclasses import dataclass, field
from typing import Optional


@dataclass
class ModelArguments:
    """
    Arguments pertaining to which model/config/tokenizer we are going to fine-tune from.
    """

    model_name_or_path: str = field(
        metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models"}
    )
    config_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained config name or path if not the same as model_name"}
    )
    tokenizer_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained tokenizer name or path if not the same as model_name"}
    )
    cache_dir: Optional[str] = field(
        default=None,
        metadata={"help": "Where to store the pretrained models downloaded from huggingface.co"},
    )
    use_fast_tokenizer: bool = field(
        default=False,
        metadata={"help": "Whether to use one of the fast tokenizer (backed by the tokenizers library) or not."},
    )
    model_revision: str = field(
        default="main",
        metadata={"help": "The specific model version to use (can be a branch name, tag name or commit id)."},
    )
    use_auth_token: bool = field(
        default=False,
        metadata={
            "help": "Will use the token generated when running `transformers-cli login` (necessary to use this script "
                    "with private models)."
        },
    )
    from_checkpoint: bool = field(
        default=False, metadata={"help": "Whether load from checkpoint to continue learning"}
    )
    model_ckpt_path: str = field(
            default=None, metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models"}
        )

@dataclass
class DataTrainingArguments:
    """
    Arguments pertaining to what data we are going to input our model for training and eval.
    """

    task: str = field(
        default="summarization",
        metadata={
            "help": "The name of the task, should be summarization (or summarization_{dataset} for evaluating "
                    "pegasus) or translation (or translation_{xx}_to_{yy})."
        },
    )
    dataset_name: Optional[str] = field(
        default=None, metadata={"help": "The name of the dataset to use (via the datasets library)."}
    )
    dataset_config_name: Optional[str] = field(
        default=None, metadata={"help": "The configuration name of the dataset to use (via the datasets library)."}
    )
    text_column: Optional[str] = field(
        default='text',
        metadata={"help": "The name of the column in the datasets containing the full texts (for summarization)."},
    )
    record_column: Optional[str] = field(
        default='record',
        metadata={"help": "The name of the column in the datasets containing the summaries (for summarization)."},
    )
    train_file: Optional[str] = field(
        default=None, metadata={"help": "The input training data file (a jsonlines or csv file)."}
    )
    validation_file: Optional[str] = field(
        default=None,
        metadata={
            "help": "An optional input evaluation data file to evaluate the metrics (rouge/sacreblue) on "
                    "(a jsonlines or csv file)."
        },
    )
    test_file: Optional[str] = field(
        default=None,
        metadata={
            "help": "An optional input test data file to evaluate the metrics (rouge/sacreblue) on "
                    "(a jsonlines or csv file)."
        },
    )
    overwrite_cache: bool = field(
        default=False, metadata={"help": "Overwrite the cached training and evaluation sets"}
    )
    preprocessing_num_workers: Optional[int] = field(
        default=None,
        metadata={"help": "The number of processes to use for the preprocessing."},
    )
    preprocess: bool = field(
        default=True,
        metadata={"help": "The number of processes to use for the preprocessing."},
    )
    preprocessed_folder: Optional[str] = field(
        default=None,
        metadata={
            "help": "Folder to preprocessed data"
        },
    )
    max_source_length: Optional[int] = field(
        default=1024,
        metadata={
            "help": "The maximum total input sequence length after tokenization. Sequences longer "
                    "than this will be truncated, sequences shorter will be padded."
        },
    )
    max_target_length: Optional[int] = field(
        default=128,
        metadata={
            "help": "The maximum total sequence length for target text after tokenization. Sequences longer "
                    "than this will be truncated, sequences shorter will be padded."
        },
    )
    max_prefix_length: Optional[int] = field(
        default=None,
        metadata={
            "help": "The maximum prefix length."
        },
    )
    val_max_target_length: Optional[int] = field(
        default=None,
        metadata={
            "help": "The maximum total sequence length for validation target text after tokenization. Sequences longer "
                    "than this will be truncated, sequences shorter will be padded. Will default to `max_target_length`."
                    "This argument is also used to override the ``max_length`` param of ``model.generate``, which is used "
                    "during ``evaluate`` and ``predict``."
        },
    )
    pad_to_max_length: bool = field(
        default=False,
        metadata={
            "help": "Whether to pad all samples to model maximum sentence length. "
                    "If False, will pad the samples dynamically when batching to the maximum length in the batch. More "
                    "efficient on GPU but very bad for TPU."
        },
    )
    max_train_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": "For debugging purposes or quicker training, truncate the number of training examples to this "
                    "value if set."
        },
    )
    max_val_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": "For debugging purposes or quicker training, truncate the number of validation examples to this "
                    "value if set."
        },
    )
    max_test_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": "For debugging purposes or quicker training, truncate the number of test examples to this "
                    "value if set."
        },
    )
    num_beams: Optional[int] = field(
        default=None,
        metadata={
            "help": "Number of beams to use for evaluation. This argument will be passed to ``model.generate``, "
                    "which is used during ``evaluate`` and ``predict``."
        },
    )
    ignore_pad_token_for_loss: bool = field(
        default=True,
        metadata={
            "help": "Whether to ignore the tokens corresponding to padded labels in the loss computation or not."
        },
    )
    source_prefix: Optional[str] = field(
        default=None, metadata={"help": "A prefix to add before every source text (useful for T5 models)."}
    )
    meta_negative: int = field(
        default=-1, metadata={"help": "Negative Schema Number in Training."}
    )
    ordered_prompt: bool = field(
        default=True,
        metadata={
            "help": "Whether to sort the spot prompt and asoc prompt or not."
        },
    )

    def __post_init__(self):
        if self.dataset_name is None and self.train_file is None and self.validation_file is None:
            raise ValueError("Need either a dataset name or a training/validation file.")
        else:
            if self.train_file is not None:
                extension = self.train_file.split(".")[-1]
                assert extension in ["csv", "json"], "`train_file` should be a csv or a json file."
            if self.validation_file is not None:
                extension = self.validation_file.split(".")[-1]
                assert extension in ["csv", "json"], "`validation_file` should be a csv or a json file."
        if self.val_max_target_length is None:
            self.val_max_target_length = self.max_target_length

    decoding_format: str = field(
        default='spotasoc',
        metadata={"help": "Decoding Format"}
    )
    record_schema: str = field(
        default=None, metadata={"help": "The input event schema file."}
    )
    spot_noise: float = field(
        default=0., metadata={"help": "The noise rate of null spot."}
    )
    asoc_noise: float = field(
        default=0., metadata={"help": "The noise rate of null asoc."}
    )
    meta_positive_rate: float = field(
        default=1., metadata={"help": "The keep rate of positive spot."}
    )
    eval_subset: str = field(default='validation')
    

@dataclass
class PromptArguments:
    use_prefix: bool = field(
        default=False,
        metadata={"help": "use prompt."},
    )
    freeze_plm: bool = field(
        default=False,
        metadata={"help": "freeze plm."},
    )
    prompt_len: int = field(
        default=80,
        metadata={"help": "prompt len."},
    )
    prompt_dim: int = field(
        default=800,
        metadata={"help": "prompt dim."},
    )
    source_prefix_path: str = field(
        default=None,
        metadata={"help": "source prefix path"},
    )
    target_prefix_path: str = field(
        default=None,
        metadata={"help": "target prefix path"},
    )
    prefix_fusion_way: str = field(
        default='add',
        metadata={"help": "[add, concat, concat_lstm]"},
    )
    freeze_prefix: bool = field(
        default=False,
        metadata={"help": "freeze prefix."},
    )
    use_adapter: bool = field(
        default=False,
        metadata={"help": "whether use adapter."},
    )
    save_prefix: bool = field(
        default=False,
        metadata={"help": "whether save prefix."},
    )
    save_label_word: bool = field(
        default=False,
        metadata={"help": "whether save label word."},
    )
    multi_source_path: str = field(
        default=None,
        metadata={"help": "the path of multi source."},
    )




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re


def fix_unk_from_text(span, text, unk='<unk>'):
    """
    Find span from the text to fix unk in the generated span
    从 text 中找到 span，修复span

    Example:
    span = "<unk> colo e Bengo"
    text = "At 159 meters above sea level , Angola International Airport is located at Ícolo e Bengo , part of Luanda Province , in Angola ."

    span = "<unk> colo e Bengo"
    text = "Ícolo e Bengo , part of Luanda Province , in Angola ."

    span = "Arr<unk> s negre"
    text = "The main ingredients of Arròs negre , which is from Spain , are white rice , cuttlefish or squid , cephalopod ink , cubanelle and cubanelle peppers . Arròs negre is from the Catalonia region ."

    span = "colo <unk>"
    text = "At 159 meters above sea level , Angola International Airport is located at e Bengo , part of Luanda Province , in Angola . coloÍ"

    span = "Tarō As<unk>"
    text = "The leader of Japan is Tarō Asō ."

    span = "Tar<unk> As<unk>"
    text = "The leader of Japan is Tarō Asō ."

    span = "<unk>Tar As<unk>"
    text = "The leader of Japan is ōTar Asō ."
    """
    if unk not in span:
        return span

    def clean_wildcard(x):
        sp = ".*?()[]+"
        return re.sub("("+"|".join([f"\\{s}" for s in sp])+")", "\\\\\g<1>", x)

    match = r'\s*\S+\s*'.join([clean_wildcard(item.strip()) for item in span.split(unk)])

    result = re.search(match, text)

    if not result:
        return span
    return result.group().strip()


def test_fix_unk_from_text():

    span_text_list = [
        ("<unk> colo e Bengo",
         "At 159 meters above sea level , Angola International Airport is located at Ícolo e Bengo , part of Luanda Province , in Angola .",
         "Ícolo e Bengo"),
        ("<unk> colo e Bengo",
         "Ícolo e Bengo , part of Luanda Province , in Angola .",
         "Ícolo e Bengo"),
        ("Arr<unk> s negre",
         "The main ingredients of Arròs negre , which is from Spain , are white rice , cuttlefish or squid , cephalopod ink , cubanelle and cubanelle peppers . Arròs negre is from the Catalonia region .",
         "Arròs negre"),
        ("colo <unk>",
         "At 159 meters above sea level , Angola International Airport is located at e Bengo , part of Luanda Province , in Angola . coloÍ",
         "coloÍ"),
        ("Tarō As<unk>", "The leader of Japan is Tarō Asō .", "Tarō Asō"),
        ("Tar<unk> As<unk>", "The leader of Japan is Tarō Asō .", "Tarō Asō"),
        ("<unk>Tar As<unk>", "The leader of Japan is ōTar Asō .", "ōTar Asō"),
        ("Atatürk Monument ( <unk> zmir )",
         "The Atatürk Monument ( İzmir ) can be found in Turkey .",
         "Atatürk Monument ( İzmir )"),
        ("The Atatürk Monument [ <unk> zmir ]",
         "The Atatürk Monument [ İzmir ] can be found in Turkey .",
         "The Atatürk Monument [ İzmir ]")
    ]

    for span, text, gold in span_text_list:
        print(span, '|', fix_unk_from_text(span, text))
        assert fix_unk_from_text(span, text) == gold


if __name__ == "__main__":
    test_fix_unk_from_text()



#!/usr/bin/env python
# coding=utf-8
import torch
from dataclasses import dataclass
from typing import Dict, Optional, Union
import numpy as np
from transformers import (
    PreTrainedTokenizerBase,
    PreTrainedModel,
)
from transformers.file_utils import PaddingStrategy


# Modified by https://github.com/huggingface/transformers/blob/2e4082364e4bd001f7933d81b3f75548704f79d7/examples/flax/language-modeling/run_t5_mlm_flax.py#L212
@dataclass
class DataCollatorForT5MLM:
    """
    Data collator used for T5 span-masked language modeling.
    It is made sure that after masking the inputs are of length `data_args.max_seq_length` and targets are also of fixed length.
    For more information on how T5 span-masked language modeling works, one can take a look
    at the `official paper <https://arxiv.org/pdf/1910.10683.pdf>`__
    or the `official code for preprocessing <https://github.com/google-research/text-to-text-transfer-transformer/blob/master/t5/data/preprocessors.py>`__ .
    Args:
        tokenizer (:class:`~transformers.PreTrainedTokenizer` or :class:`~transformers.PreTrainedTokenizerFast`):
            The tokenizer used for encoding the data.
        noise_density (:obj:`float`):
            The probability with which to (randomly) mask tokens in the input.
        mean_noise_span_length (:obj:`float`):
            The average span length of the masked tokens.
        input_length (:obj:`int`):
            The expected input length after masking.
        target_length (:obj:`int`):
            The expected target length after masking.
        pad_token_id: (:obj:`int`):
            The pad token id of the model
        decoder_start_token_id: (:obj:`int):
            The decoder start token id of the model
    """

    tokenizer: PreTrainedTokenizerBase
    model: Optional[PreTrainedModel]
    pad_token_id: int
    decoder_start_token_id: int
    noise_density: float = 0.15
    mean_noise_span_length: float = 3
    max_length: Optional[int] = None
    pad_to_multiple_of: Optional[int] = None
    padding: Union[bool, str, PaddingStrategy] = True

    def __call__(self, features) -> Dict[str, np.ndarray]:
        """ Make T5 MLM Batch

        Args:
            features (List[Dict[str, List]]):
                - input_ids
                - attention_mask

        Returns:
            [type]: [description]
        """
        raw_input_ids = [feature["input_ids"] for feature in features]
        input_ids = self.tokenizer.pad(
            [{'input_ids': x} for x in raw_input_ids],
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="np",
            return_attention_mask=False,
        )
        input_ids = input_ids['input_ids']

        mask_indices = np.zeros_like(input_ids, dtype=bool)
        for index, x in enumerate(raw_input_ids):
            if len(x) - 1 < 2:
                # Ignore src with single token
                continue
            mask = self.random_spans_noise_mask(len(x) - 1)
            mask_indices[index, :len(mask)] = mask

        labels_mask = ~mask_indices

        input_ids_sentinel = self.create_sentinel_ids(mask_indices.astype(np.int8))
        labels_sentinel = self.create_sentinel_ids(labels_mask.astype(np.int8))
        features = {
            'input_ids': torch.tensor(self.filter_input_ids(input_ids, input_ids_sentinel)),
            'labels': torch.tensor(self.filter_input_ids(input_ids, labels_sentinel)).contiguous(),
        }
        features['attention_mask'] = features['input_ids'] > 0

        # prepare decoder_input_ids
        if self.model is not None and hasattr(self.model, "prepare_decoder_input_ids_from_labels"):
            decoder_input_ids = self.model.prepare_decoder_input_ids_from_labels(labels=features["labels"])
            features["decoder_input_ids"] = decoder_input_ids

        return features

    def create_sentinel_ids(self, mask_indices):
        """
        Sentinel ids creation given the indices that should be masked.
        The start indices of each mask are replaced by the sentinel ids in increasing
        order. Consecutive mask indices to be deleted are replaced with `-1`.
        """
        start_indices = mask_indices - np.roll(mask_indices, 1, axis=-1) * mask_indices
        start_indices[:, 0] = mask_indices[:, 0]

        sentinel_ids = np.where(start_indices != 0, np.cumsum(start_indices, axis=-1), start_indices)
        sentinel_ids = np.where(sentinel_ids != 0, (sentinel_ids + self.tokenizer.vocab_size - 101), 0)
        sentinel_ids -= mask_indices - start_indices

        return sentinel_ids

    def filter_input_ids(self, input_ids, sentinel_ids):
        """
        Puts sentinel mask on `input_ids` and fuse consecutive mask tokens into a single mask token by deleting.
        This will reduce the sequence length from `expanded_inputs_length` to `input_length`.
        """
        new_input_ids = list()
        for x, y in zip(input_ids, sentinel_ids):
            input_ids_full = np.where(y != 0, y, x)
            x = input_ids_full[input_ids_full > 0].tolist()[:-1]
            x = x + [self.tokenizer.eos_token_id]
            new_input_ids += [x]
        from itertools import zip_longest

        new_input_ids = np.array(list(zip_longest(*new_input_ids, fillvalue=0))).T
        return new_input_ids

    def random_spans_noise_mask(self, length):

        """This function is copy of `random_spans_helper <https://github.com/google-research/text-to-text-transfer-transformer/blob/84f8bcc14b5f2c03de51bd3587609ba8f6bbd1cd/t5/data/preprocessors.py#L2682>`__ .
        Noise mask consisting of random spans of noise tokens.
        The number of noise tokens and the number of noise spans and non-noise spans
        are determined deterministically as follows:
        num_noise_tokens = round(length * noise_density)
        num_nonnoise_spans = num_noise_spans = round(num_noise_tokens / mean_noise_span_length)
        Spans alternate between non-noise and noise, beginning with non-noise.
        Subject to the above restrictions, all masks are equally likely.
        Args:
            length: an int32 scalar (length of the incoming token sequence)
            noise_density: a float - approximate density of output mask
            mean_noise_span_length: a number
        Returns:
            a boolean tensor with shape [length]
        """
        orig_length = length

        num_noise_tokens = int(np.round(length * self.noise_density))
        # avoid degeneracy by ensuring positive numbers of noise and nonnoise tokens.
        num_noise_tokens = min(max(num_noise_tokens, 1), length - 1)
        num_noise_spans = int(np.round(num_noise_tokens / self.mean_noise_span_length))
        # avoid degeneracy by ensuring positive number of noise spans
        num_noise_spans = max(num_noise_spans, 1)
        num_nonnoise_tokens = length - num_noise_tokens

        # pick the lengths of the noise spans and the non-noise spans
        def _random_segmentation(num_items, num_segments):
            """Partition a sequence of items randomly into non-empty segments.
            Args:
                num_items: an integer scalar > 0
                num_segments: an integer scalar in [1, num_items]
            Returns:
                a Tensor with shape [num_segments] containing positive integers that add
                up to num_items
            """
            mask_indices = np.arange(num_items - 1) < (num_segments - 1)
            np.random.shuffle(mask_indices)
            first_in_segment = np.pad(mask_indices, [[1, 0]])
            segment_id = np.cumsum(first_in_segment)
            # count length of sub segments assuming that list is sorted
            _, segment_length = np.unique(segment_id, return_counts=True)
            return segment_length

        noise_span_lengths = _random_segmentation(num_noise_tokens, num_noise_spans)
        nonnoise_span_lengths = _random_segmentation(num_nonnoise_tokens, num_noise_spans)

        interleaved_span_lengths = np.reshape(
            np.stack([nonnoise_span_lengths, noise_span_lengths], axis=1), [num_noise_spans * 2]
        )
        span_starts = np.cumsum(interleaved_span_lengths)[:-1]
        span_start_indicator = np.zeros((length,), dtype=np.int8)
        span_start_indicator[span_starts] = True
        span_num = np.cumsum(span_start_indicator)
        is_noise = np.equal(span_num % 2, 1)

        return is_noise[:orig_length]



from transformers import AutoTokenizer
from torch.nn.utils.rnn import pad_sequence
import torch
import json

def split_label_words(tokenizer, label_list):
    label_word_list = []
    for label in label_list:
        if label == 'no_relation':
            label_word_id = tokenizer.encode('None', add_special_tokens=False)
            label_word_list.append(torch.tensor(label_word_id))
        else:
            tmps = label
            label = label.lower()
            label = label.split("(")[0]
            label = label.replace(":"," ").replace("_"," ").replace("per","person").replace("org","organization")
            label_word_id = tokenizer(label, add_special_tokens=False)['input_ids']
            print(label, label_word_id)
            label_word_list.append(torch.tensor(label_word_id))
    padded_label_word_list = pad_sequence([x for x in label_word_list], batch_first=True, padding_value=0)
    return padded_label_word_list


def get_label_word(cfg):
    model_name_or_path = cfg.model_name_or_path

    tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
    with open("data/rel2id.json", "r") as file:
        t = json.load(file)
        label_list = list(t)

    t = split_label_words(tokenizer, label_list)

    with open(f"data/{model_name_or_path}.pt", "wb") as file:
        torch.save(t, file)




import os
import numpy as np
import json
import shutil




def get_labels(path, name,  negative_label="no_relation"):
    """See base class."""

    with open(path + "/" + name, "r") as f:
        features = []
        for line in f.readlines():
            line = line.rstrip()
            if len(line) > 0:
                features.append(eval(line))
    return features

def generate_k_shot(data_dir):
    Seed = [1, 2, 3, 4, 5]
    mode = 'k-shot'
    data_file = 'train.txt'
    path = 'data'

    output_dir = os.path.join(path, mode)
    dataset = get_labels(path, data_file)

    for seed in Seed:

        # Other datasets
        np.random.seed(seed)
        np.random.shuffle(dataset)

        # Set up dir
        k = 8
        setting_dir = os.path.join(output_dir, f"{k}-{seed}")
        os.makedirs(setting_dir, exist_ok=True)

        label_list = {}
        for line in dataset:
            label = line['relation']
            if label not in label_list:
                label_list[label] = [line]
            else:
                label_list[label].append(line)

        with open(os.path.join(setting_dir, "train.txt"), "w") as f:
            file_list = []
            for label in label_list:
                for line in label_list[label][:k]:  # train中每一类取前k个数据
                    f.writelines(json.dumps(line))
                    f.write('\n')

            f.close()

    shutil.copyfile('data/rel2id.json','data/k-shot/8-1/rel2id.json')
    shutil.copyfile('data/val.txt','data/k-shot/8-1/val.txt')
    shutil.copyfile('data/test.txt','data/k-shot/8-1/test.txt')



import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../")))
from dataclasses import dataclass
from typing import Optional, Tuple, Any, Dict, List
import torch

from transformers.file_utils import (
    add_code_sample_docstrings,
    add_end_docstrings,
    add_start_docstrings,
    add_start_docstrings_to_model_forward,
    replace_return_docstrings,
    ModelOutput
)
from transformers.modeling_outputs import (
    Seq2SeqLMOutput
)

@dataclass
class MySeq2SeqLMOutput(Seq2SeqLMOutput):
    """
    Extra sequence-to-sequence language models outputs.
    Args:
        pairing (`torch.LongTensor` of shape `(1,)`, *optional*):
            pairing left brackets for right brackets.
        linking (`torch.LongTensor` of shape `(1,)`, *optional*):
            linking right brackets to right brackets.
        typing (`torch.LongTensor` of shape `(1,)`, *optional*):
            typing right brackets.
    """    
    pairing: Optional[List[torch.LongTensor]] = None
    linking: Optional[List[torch.LongTensor]] = None
    typing: Optional[List[torch.LongTensor]] = None



#!/usr/bin/env python
# -*- coding:utf-8 -*-

spot_prompt = '<spot>'
asoc_prompt = '<asoc>'

type_start = '<extra_id_0>'
type_end = '<extra_id_1>'
text_start = '<extra_id_2>'
span_start = '<extra_id_5>'
null_span = '<extra_id_6>'
null_label = '<extra_id_7>'


class StructureMarker:
    def __init__(self) -> None:
        pass


class BaseStructureMarker(StructureMarker):
    def __init__(self) -> None:
        super().__init__()
        self.sent_start = '<extra_id_0>'
        self.sent_end = '<extra_id_1>'
        self.record_start = '<extra_id_0>'
        self.record_end = '<extra_id_1>'
        self.span_start = '<extra_id_0>'
        self.span_end = '<extra_id_1>'
        self.text_start = '<extra_id_2>'
        self.source_span_start = '<extra_id_3>'
        self.source_span_end = '<extra_id_4>'
        self.target_span_start = '<extra_id_5>'
        self.null_span = '<extra_id_6>'
        self.null_label = '<extra_id_7>'



import torch
import numpy as np
import random
from torch import nn
import torch.nn.functional as F
from transformers import BartModel, BartTokenizer, BertTokenizer
import os

def avg_token_embeddings(tokenizer: BartTokenizer, bart_model: BartModel, bart_name, num_tokens):
    """when initial added tokens, use their averge token emebddings

    Args:
        tokenizer (BartTokenizer): [description]
        bart_model (BartModel): [description]
        bart_name ([type]): [description]
        num_tokens ([type]): [description]

    Raises:
        RuntimeError: [description]

    Returns:
        [type]: [description]
    """
    _tokenizer = BertTokenizer.from_pretrained(bart_name) if 'chinese' in bart_name \
        else BartTokenizer.from_pretrained(bart_name)
    for token in tokenizer.unique_no_split_tokens:
        if token[:2] == '<<':  # 特殊字符
            index = tokenizer.convert_tokens_to_ids(tokenizer.tokenize(token))
            if len(index)>1:
                raise RuntimeError(f"{token} wrong split")
            else:
                index = index[0]
            assert index>=num_tokens, (index, num_tokens, token)
            indexes = _tokenizer.convert_tokens_to_ids(_tokenizer.tokenize(token[2:-2]))
            embed = bart_model.encoder.embed_tokens.weight.data[indexes[0]]
            for i in indexes[1:]:
                embed += bart_model.decoder.embed_tokens.weight.data[i]
            embed /= len(indexes)
            bart_model.decoder.embed_tokens.weight.data[index] = embed
    return bart_model


def seq_to_mask(seq_len, max_len):
    """[get attention mask with sequence length]

    Args:
        seq_len ([torch.tensor]): [shape: bsz, each sequence length in a batch]
    """
    max_len = int(max_len) if max_len else seq_len.max().long()
    cast_seq = torch.arange(max_len).expand(seq_len.size(0), -1).to(seq_len)
    mask = cast_seq.lt(seq_len.unsqueeze(1))
    return mask

def get_loss(tgt_tokens, tgt_seq_len, pred):
    """

    :param tgt_tokens: bsz x max_len, 包含了的[sos, token, eos]
    :param pred: bsz x max_len-1 x vocab_size
    :return:
    """
    tgt_seq_len = tgt_seq_len - 1
    mask = seq_to_mask(tgt_seq_len, max_len=tgt_tokens.size(1) - 1).eq(0)
    tgt_tokens = tgt_tokens[:, 1:].masked_fill(mask, -100)
    loss = F.cross_entropy(target=tgt_tokens, input=pred.transpose(1, 2))
    return loss

def get_model_device(model):
    assert isinstance(model, nn.Module)

    parameters = list(model.parameters())
    if len(parameters) == 0:
        return None
    else:
        return parameters[0].device

def set_seed(seed=2021):
    """sets random seed"""
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    np.random.seed(seed)
    random.seed(seed)


def convert_preds_to_outputs(preds, raw_words, mapping, tokenizer):
    """convet model predicitons to BIO outputs

    Args:
        preds ([torch.Tensor]): [prompt model predictions, (bsz x seq_len x labels)]
        raw_words ([List]): [source raw words]
        mapping ([dict]): [map entity labels to <<>>]
        tokenizer : [BartTokenizer]

    Returns:
        [outputs (List)]: [each item length equal to raw_words, BIO format.]
    """
    id2label = list(mapping.keys())
    pred_eos_index = preds.flip(dims=[1]).eq(1).cumsum(dim=1).long()
    preds = preds[:, 1:]
    pred_seq_len = pred_eos_index.flip(dims=[1]).eq(pred_eos_index[:, -1:]).sum(dim=1)  # bsz
    pred_seq_len = (pred_seq_len - 2).tolist()

    word_start_index = len(mapping) + 2
    outputs = []
    for i, pred_item in enumerate(preds.tolist()):
        pred_item = pred_item[:pred_seq_len[i]] # single sentence prediction
        pairs, cur_pair = [], []
        if len(pred_item):  # this sentence prediciton= is not null
            for idx in pred_item:
                if idx < word_start_index:  # is entity
                    if len(cur_pair) > 0:
                        # assert word[i] < word[i+1]
                        if all([cur_pair[i] < cur_pair[i + 1] for i in range(len(cur_pair) - 1)]):
                            pairs.append(tuple(cur_pair + [idx]))   # add valid words and current entity id
                    cur_pair = []   # clear word pairs
                else:   # is word
                    cur_pair.append(idx)    # add word id to word pairs
        raw_words_item = raw_words[i]
        cum_lens = [1]
        start_idx = 1
        for word in raw_words_item:
            start_idx += len(tokenizer.tokenize(word, add_prefix_space=True))
            cum_lens.append(start_idx)
        cum_lens.append(start_idx+1)
        output = ['O' for _ in range(len(raw_words_item))]
        # pairs: List[(word id, ... , entity id), (...), ...]
        for pair in pairs:  # (word id, ... , entity id)
            entity = pair[-1]
            words = []
            for word in pair[:-1]:
                if word-word_start_index in cum_lens:
                    words.append(cum_lens.index(word-word_start_index)) 
            if len(words) == 0: continue
            start_idx = words[0]
            end_idx = words[-1]
            output[start_idx] = f'B-{id2label[entity-2]}'
            for _ in range(start_idx+1, end_idx+1):
                output[_] = f'I-{id2label[entity-2]}'
        outputs.append(output)
    return outputs


def write_predictions(path, texts, labels):
    """[write model predictions to path (conll format)]

    Args:
        path ([str]): [save path]
        texts ([List]): [raw texts]
        labels ([List]): [predict labels]
    """
    print(len(texts), len(labels))
    assert len(texts) == len(labels)
    if not os.path.exists(path):
        os.system(r"touch {}".format(path))
    
    with open(path, "w", encoding="utf-8") as f:
        f.writelines("-DOCSTART-	O\n\n")
        for i in range(len(texts)):
            for j in range(len(texts[i])):
                f.writelines("{}\t{}\n".format(texts[i][j], labels[i][j]))
            f.writelines("\n")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import torch
import torch.nn as nn
from dataclasses import dataclass, field
from typing import Union, List, Dict, Tuple, Any, Optional
from torch.cuda.amp import autocast

from transformers import (
    Seq2SeqTrainer,
    Seq2SeqTrainingArguments, )
from transformers.trainer_pt_utils import LabelSmoother

from transformers.trainer import *

from ..seq2seq.constraint_decoder import get_constraint_decoder


@dataclass
class ConstraintSeq2SeqTrainingArguments(Seq2SeqTrainingArguments):
    """
    Parameters:
        constraint_decoding (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Whether to use Constraint Decoding
        structure_weight (:obj:`float`, `optional`, defaults to :obj:`None`):
    """
    constraint_decoding: bool = field(default=False, metadata={"help": "Whether to Constraint Decoding or not."})
    save_better_checkpoint: bool = field(default=False,
                                         metadata={"help": "Whether to save better metric checkpoint"})
    start_eval_step: int = field(default=0, metadata={"help": "Start Evaluation after Eval Step"})


class ConstraintSeq2SeqTrainer(Seq2SeqTrainer):
    def __init__(self, decoding_type_schema=None, task='event', decoding_format='tree', source_prefix=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.decoding_format = decoding_format
        self.decoding_type_schema = decoding_type_schema

        # Label smoothing by sum token loss, different from different Label smootheing
        if self.args.label_smoothing_factor != 0:
            self.label_smoother = LabelSmoother(epsilon=self.args.label_smoothing_factor)
            print('Using %s' % self.label_smoother)
        else:
            self.label_smoother = None

        if self.args.constraint_decoding:
            self.constraint_decoder = get_constraint_decoder(tokenizer=self.tokenizer,
                                                             type_schema=self.decoding_type_schema,
                                                             decoding_schema=self.decoding_format,
                                                             source_prefix=source_prefix,
                                                             task_name=task)
        else:
            self.constraint_decoder = None

        self.oom_batch = 0

    def training_step(self, model: nn.Module, inputs: Dict[str, Union[torch.Tensor, Any]]) -> torch.Tensor:
        """
        Perform a training step on a batch of inputs.

        Subclass and override to inject custom behavior.

        Args:
            model (:obj:`nn.Module`):
                The model to train.
            inputs (:obj:`Dict[str, Union[torch.Tensor, Any]]`):
                The inputs and targets of the model.

                The dictionary will be unpacked before being fed to the model. Most models expect the targets under the
                argument :obj:`labels`. Check your model's documentation for all accepted arguments.

        Return:
            :obj:`torch.Tensor`: The tensor with training loss on this batch.
        """

        oom = False
        oom_message = ""
        try:
            loss = super().training_step(model, inputs)
            return loss
        except RuntimeError as e:
            if 'out of memory' in str(e):
                oom = True
                oom_message = str(e)
                logger.warning(f'ran out of memory {self.oom_batch} on {self.args.local_rank}')
                for k, v in inputs.items():
                    print(k, v.size())
            else:
                raise e

        if oom:
            self.oom_batch += 1
            raise RuntimeError(oom_message)

    def train(
        self,
        resume_from_checkpoint: Optional[Union[str, bool]] = None,
        trial: Union["optuna.Trial", Dict[str, Any]] = None,
        **kwargs,
    ):
        return super().train(
            resume_from_checkpoint=resume_from_checkpoint,
            trial=trial,
            **kwargs
        )

    def _maybe_log_save_evaluate(self, tr_loss, model, trial, epoch, ignore_keys_for_eval):
        if self.control.should_log:
            logs: Dict[str, float] = {}
            tr_loss_scalar = tr_loss.item()
            # reset tr_loss to zero
            tr_loss -= tr_loss

            logs["loss"] = round(tr_loss_scalar / (self.state.global_step - self._globalstep_last_logged), 4)
            logs["learning_rate"] = self._get_learning_rate()

            self._total_loss_scalar += tr_loss_scalar
            self._globalstep_last_logged = self.state.global_step

            self.log(logs)

        if self.args.start_eval_step > 0 and self.state.global_step < self.args.start_eval_step:
            return

        previous_best_metric = self.state.best_metric
        metrics = None
        if self.control.should_evaluate:
            metrics = self.evaluate()
            self._report_to_hp_search(trial, epoch, metrics)

        # Only save the checkpoint better than previous_best_metric
        if self.args.save_better_checkpoint and self.args.metric_for_best_model is not None:
            if metrics is not None and previous_best_metric is not None:
                if metrics[self.args.metric_for_best_model] <= previous_best_metric:
                    return

        if self.control.should_save:
            self._save_checkpoint(model, trial, metrics=metrics)
            self.control = self.callback_handler.on_save(self.args, self.state, self.control)

    def prediction_step(
            self,
            model: nn.Module,
            inputs: Dict[str, Union[torch.Tensor, Any]],
            prediction_loss_only: bool,
            ignore_keys: Optional[List[str]] = None,
    ) -> Tuple[Optional[float], Optional[torch.Tensor], Optional[torch.Tensor]]:
        """
        Perform an evaluation step on :obj:`model` using obj:`inputs`.

        Subclass and override to inject custom behavior.

        Args:
            model (:obj:`nn.Module`):
                The model to evaluate.
            inputs (:obj:`Dict[str, Union[torch.Tensor, Any]]`):
                The inputs and targets of the model.

                The dictionary will be unpacked before being fed to the model. Most models expect the targets under the
                argument :obj:`labels`. Check your model's documentation for all accepted arguments.
            prediction_loss_only (:obj:`bool`):
                Whether or not to return the loss only.

        Return:
            Tuple[Optional[float], Optional[torch.Tensor], Optional[torch.Tensor]]: A tuple with the loss, logits and
            labels (each being optional).
        """

        def prefix_allowed_tokens_fn(batch_id, sent):
            # print(self.tokenizer.convert_ids_to_tokens(inputs['labels'][batch_id]))
            src_sentence = inputs['input_ids'][batch_id]
            return self.constraint_decoder.constraint_decoding(src_sentence=src_sentence,
                                                               tgt_generated=sent)
 
        if not self.args.predict_with_generate or prediction_loss_only:
            return super().prediction_step(
                model=model,
                inputs=inputs,
                prediction_loss_only=prediction_loss_only,
                ignore_keys=ignore_keys,
            )
        has_labels = "labels" in inputs
        inputs = self._prepare_inputs(inputs)

        gen_kwargs = {
            "max_length": self.model.config.max_length,
            "num_beams": self.model.config.num_beams,
            "generation_config": self.model.t5.generation_config,   # modify by xxx
            "prefix_allowed_tokens_fn": prefix_allowed_tokens_fn if self.constraint_decoder else None,
        }

        generated_tokens = self.model.generate(
            inputs["input_ids"],
            attention_mask=inputs["attention_mask"],
            labels=None,
            **gen_kwargs,
        )

        # in case the batch is shorter than max length, the output should be padded
        if generated_tokens.shape[-1] < gen_kwargs["max_length"]:
            generated_tokens = self._pad_tensors_to_max_len(generated_tokens, gen_kwargs["max_length"])

        with torch.no_grad():
            outputs = model(**inputs)
            if has_labels:
                if self.label_smoother is not None:
                    loss = self.label_smoother(outputs, inputs["labels"]).mean().detach()
                else:
                    loss = (outputs["loss"] if isinstance(outputs, dict) else outputs[0]).mean().detach()
            else:
                loss = None

        if self.args.prediction_loss_only:
            return loss, None, None

        labels = inputs["labels"]
        if labels.shape[-1] < gen_kwargs["max_length"]:
            labels = self._pad_tensors_to_max_len(labels, gen_kwargs["max_length"])

        return loss, generated_tokens, labels

    def save_model(self, output_dir: Optional[str] = None, _internal_call: bool = False):
        """
        Will save the model, so you can reload it using `from_pretrained()`.

        Will only save from the main process.
        """
        if output_dir is None:
            output_dir = self.args.output_dir
        
        self._save(output_dir)
    
    def _save(self, output_dir: Optional[str] = None, state_dict=None):
        # If we are executing this function, we are the process zero, so we don't check for that.
        output_dir = output_dir if output_dir is not None else self.args.output_dir
        os.makedirs(output_dir, exist_ok=True)
        logger.info(f"Saving model checkpoint to {output_dir}")
        # Save a trained model and configuration using `save_pretrained()`.
        # They can then be reloaded using `from_pretrained()`
        if not isinstance(self.model, PreTrainedModel):
            if isinstance(unwrap_model(self.model), PreTrainedModel):
                if state_dict is None:
                    state_dict = self.model.state_dict()
                unwrap_model(self.model).save_pretrained(output_dir, state_dict=state_dict)
            else:
                logger.info("Trainer.model is not a `PreTrainedModel`, only saving its state dict.")
                if state_dict is None:
                    state_dict = self.model.state_dict()
                torch.save(state_dict, os.path.join(output_dir, WEIGHTS_NAME))
        else:
            self.model.save_pretrained(output_dir, state_dict=state_dict)
        if self.tokenizer is not None:
            self.tokenizer.save_pretrained(output_dir)
            
        if self.model.config is not None:
            self.model.config.save_pretrained(output_dir)

        # Good practice: save your training arguments together with the trained model
        torch.save(self.args, os.path.join(output_dir, TRAINING_ARGS_NAME))

def main(): pass


if __name__ == "__main__":
    main()
